# -----------------------------------------------------------------------------
# Copyright (c) 2020- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Main plugin widget.

SpyderDockablePlugin plugins must provide a WIDGET_CLASS attribute that is a
subclass of PluginMainWidget.
"""

from __future__ import annotations

# Standard library imports
import logging
from collections import OrderedDict
from typing import TYPE_CHECKING

# Third party imports
from qtpy import PYSIDE2
from qtpy.QtCore import QByteArray, QSize, Qt, Signal, Slot
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QSizePolicy,
    QStackedWidget,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

# Local imports
from spyder.api.exceptions import SpyderAPIError
from spyder.api.translations import _
from spyder.api.widgets import PluginMainWidgetActions, PluginMainWidgetWidgets
from spyder.api.widgets.auxiliary_widgets import (
    MainCornerWidget,
    SpyderWindowWidget,
)
from spyder.api.widgets.menus import (
    OptionsMenuSections,
    PluginMainWidgetMenus,
    PluginMainWidgetOptionsMenu,
)
from spyder.api.widgets.mixins import SpyderWidgetMixin
from spyder.api.widgets.toolbars import MainWidgetToolbar
from spyder.utils.qthelpers import create_waitspinner, qbytearray_to_str
from spyder.utils.registries import ACTION_REGISTRY, TOOLBAR_REGISTRY
from spyder.utils.stylesheet import (
    AppStyle,
    APP_STYLESHEET,
    PANES_TABBAR_STYLESHEET,
    PANES_TOOLBAR_STYLESHEET,
)
from spyder.widgets.dock import DockTitleBar, SpyderDockWidget
from spyder.widgets.emptymessage import EmptyMessageWidget
from spyder.widgets.tabs import Tabs

if TYPE_CHECKING:
    from qtpy.QtGui import QCloseEvent, QFocusEvent
    from qtpy.QWidget import QLayout

    import spyder.app.mainwindow  # For MainWindow
    import spyder.utils.qthelpers  # For SpyderAction
    import spyder.widgets.dock  # For SpyderDockWidget

    from spyder.api.plugins import SpyderPluginV2

# Logging
logger = logging.getLogger(__name__)


class PluginMainWidget(QWidget, SpyderWidgetMixin):
    """
    Spyder plugin main widget class.

    This class handles both a dockwidget pane and a floating window widget
    (undocked pane).

    .. note::

        All :class:`~spyder.api.plugins.SpyderDockablePlugin`\\s define a
        main widget that must be a subclass of this class.

    Notes
    -----
    This widget is a subclass of :class:`QWidget` that consists of a single
    central widget and a set of toolbars stacked above it.

    The toolbars are not movable nor floatable and must occupy the entire
    horizontal space available for the plugin. This mean that toolbars must be
    stacked vertically and cannot be placed horizontally next to each other.
    """

    # ---- Attributes
    # -------------------------------------------------------------------------
    ENABLE_SPINNER: bool = False
    """
    Enable/disable showing a progress spinner on the top right of the toolbar.

    If ``True``, an extra space will be added to the toolbar (even if the
    spinner is not moving) to avoid items jumping to the left/right when
    the spinner appears. If ``False`` no extra space will be added.

    The spinner is shown to the left of the Options (hamburger) menu.

    Plugins that provide actions that take time should make this ``True`` and
    use the :meth:`start_spinner`/:meth:`stop_spinner` methods accordingly.

    Examples
    --------
    The :guilabel:`Find in Files` plugin (:mod:`spyder.plugins.findinfiles`
    is an example of a core plugin that uses it.
    """

    CONTEXT_NAME: str | None = None
    """
    The name under which to store actions, toolbars, toolbuttons and menus.

    This optional attribute defines the context name under which actions,
    toolbars, toolbuttons and menus should be registered in the
    Spyder global registry.

    If those elements belong to the global scope of the plugin, then this
    attribute should have a ``None`` value, which will use the plugin's name as
    the context scope.
    """

    MARGIN_TOP: int = 0
    """
    Adjust the widget's top margin in integer pixels.
    """

    SHOW_MESSAGE_WHEN_EMPTY: bool = False
    """
    Enable or (by default) disable showing a message when the widget is empty.

    .. note ::

        If ``True``, you need to set at least the :attr:`MESSAGE_WHEN_EMPTY`
        attribute as well.

    Examples
    --------
    The :guilabel:`Find in Files` plugin is an example of a core plugin
    that uses it.
    """

    MESSAGE_WHEN_EMPTY: str | None = None
    """
    The main message, as a string, that will be shown when the widget is empty.

    Must be set to a string  when :attr:`SHOW_MESSAGE_WHEN_EMPTY` is ``True``,
    and has no effect if that attribute is ``False``.

    Examples
    --------
    The :guilabel:`Find in Files` plugin is an example of a core plugin
    that uses it.
    """

    IMAGE_WHEN_EMPTY: str | None = None
    """
    Name of or path to an SVG image to show when the widget is empty.

    If ``None`` (the default), no image is shown.

    Only shown when :attr:`SHOW_MESSAGE_WHEN_EMPTY` is set to ``True``.

    .. note ::

        This needs to be an SVG file so that it can be rendered correctly
        on high-resolution screens.

    Examples
    --------
    The :guilabel:`Find in Files` plugin is an example of a core plugin
    that uses it.
    """

    DESCRIPTION_WHEN_EMPTY: str | None = None
    """
    Additional text shown below the main message when the widget is empty.

    If ``None`` (the default), no additional text is shown.

    Only shown when :attr:`SHOW_MESSAGE_WHEN_EMPTY` is set to ``True``,
    and shown below :attr:`MESSAGE_WHEN_EMPTY`.

    Examples
    --------
    The :guilabel:`Find in Files` plugin is an example of a core plugin
    that uses it.
    """

    SET_LAYOUT_WHEN_EMPTY: bool = True
    """
    Use a vertical layout for the stack holding the empty and content widgets.

    Set this to ``False`` if you need to use a more complex layout in
    your widget; ``True`` is the default behavior.

    Examples
    --------
    The :guilabel:`Debugger` plugin is an example of a core plugin
    that uses it.
    """

    # ---- Signals
    # -------------------------------------------------------------------------
    sig_free_memory_requested: Signal = Signal()
    """
    Signal to request the main application garbage-collect deleted objects.
    """

    sig_quit_requested: Signal = Signal()
    """
    Signal to request the main Spyder application quit.
    """

    sig_restart_requested: Signal = Signal()
    """
    Signal to request the main Spyder application quit and restart itself.
    """

    sig_redirect_stdio_requested: Signal = Signal(bool)
    """
    Request the main app redirect standard out/error within file pickers.

    This will redirect :data:`~sys.stdin`, :data:`~sys.stdout`, and
    :data:`~sys.stderr` when using :guilabel:`Open`, :guilabel:`Save`,
    and :guilabel:`Browse` dialogs within a plugin's widgets.

    Parameters
    ----------
    enable: bool
        Enable (``True``) or disable (``False``) standard input/output
        redirection.
    """

    sig_exception_occurred: Signal = Signal(dict)
    """
    Signal to report an exception from a plugin.

    Parameters
    ----------
    error_data: dict[str, str | bool]
        The dictionary containing error data. The expected keys are:

        .. code-block:: python

            error_data = {
                "text": str,
                "is_traceback": bool,
                "repo": str,
                "title": str,
                "label": str,
                "steps": str,
            }

        The ``is_traceback`` key indicates if ``text`` contains plain text or a
        Python error traceback.

        The ``title`` and ``repo`` keys indicate how the error data should
        customize the report dialog and GitHub error submission.

        The ``label`` and ``steps`` keys allow customizing the content of the
        error dialog.
    """

    sig_toggle_view_changed: Signal = Signal(bool)
    """
    Signal to report that visibility of a dockable plugin has changed.

    This is triggered by checking/unchecking the entry for a pane in the
    :menuselection:`Window --> Panes` menu.

    Parameters
    ----------
    visible: bool
        Whether the widget has been shown (``True``) or hidden (``False``).
    """

    sig_update_ancestor_requested: Signal = Signal()
    """
    Notify the main window that a child widget needs its ancestor updated.
    """

    sig_unmaximize_plugin_requested: Signal = Signal((), (object,))
    """
    Request the main window unmaximize the currently maximized plugin, if any.

    If emitted without arguments, it'll unmaximize any plugin.

    Parameters
    ----------
    plugin_instance: spyder.api.plugins.SpyderDockablePlugin
        Unmaximize current plugin only if it is not ``plugin_instance``.
    """

    sig_focus_status_changed: Signal = Signal(bool)
    """
    Signal to report a change in the focus state of this widget.

    Parameters
    ----------
    status: bool
        ``True`` if the widget is now focused; ``False`` if it is not.
    """

    def __init__(
        self,
        name: str,
        plugin: SpyderPluginV2,
        parent: spyder.app.mainwindow.MainWindow | None = None,
    ) -> None:
        """
        Create a new main widget for a plugin.

        The widget is created automatically by Spyder, and is not intended
        to be instantiated manually.

        Parameters
        ----------
        name : str
            The name of the plugin, i.e. the
            :attr:`SpyderPluginV2.NAME <spyder.api.plugins.SpyderPluginV2.NAME>`.
        plugin : SpyderPluginV2
            The plugin object this is to be the container class of.
        parent : spyder.app.mainwindow.MainWindow | None, optional
            The container's parent widget, normally the Spyder main window.
            By default (``None``), no parent widget (used for testing).

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If :attr:`SHOW_MESSAGE_WHEN_EMPTY` is set to ``True`` but
            :attr:`MESSAGE_WHEN_EMPTY` is not set to a non-empty string.
        """
        if not PYSIDE2:
            super().__init__(parent=parent, class_parent=plugin)
        else:
            QWidget.__init__(self, parent)
            SpyderWidgetMixin.__init__(self, class_parent=plugin)

        # Attributes
        # --------------------------------------------------------------------
        self._is_tab = False
        self._name = name
        self._plugin = plugin
        self._parent = parent
        self._default_margins = None
        self.is_visible: bool | None = None
        self.dock_action: spyder.utils.qthelpers.SpyderAction | None = None
        self.undock_action: spyder.utils.qthelpers.SpyderAction | None = None
        self.close_action: spyder.utils.qthelpers.SpyderAction | None = None
        self._toolbars_already_rendered = False
        self._is_maximized = False

        self.PLUGIN_NAME: str = name
        """
        Plugin name in the action, toolbar, toolbutton & menu registries.

        Usually the same as
        :attr:`SpyderPluginV2.NAME <spyder.api.plugins.SpyderPluginV2.NAME>`,
        but may be different from :attr:`CONTEXT_NAME`.
        """

        # We create our toggle action instead of using the one that comes with
        # dockwidget because it was not possible to raise and focus the plugin
        self.toggle_view_action: spyder.utils.qthelpers.SpyderAction | None = (
            None
        )
        self._toolbars = OrderedDict()
        self._auxiliary_toolbars = OrderedDict()

        # Widgets
        # --------------------------------------------------------------------
        self.windowwidget: SpyderWindowWidget | None = None
        self.dockwidget: spyder.widgets.dock.SpyderDockWidget | None = None
        self._icon = QIcon()
        self._spinner = None
        self._stack = None
        self._content_widget = None
        self._pane_empty = None

        if self.ENABLE_SPINNER:
            self._spinner = create_waitspinner(
                size=16, parent=self, name=PluginMainWidgetWidgets.Spinner
            )

        self._corner_widget = MainCornerWidget(
            parent=self,
            name=PluginMainWidgetWidgets.CornerWidget,
        )
        self._corner_widget.ID = "main_corner"

        self._main_toolbar = MainWidgetToolbar(
            parent=self,
            title=_("Main widget toolbar"),
        )
        self._main_toolbar.ID = "main_toolbar"

        TOOLBAR_REGISTRY.register_reference(
            self._main_toolbar,
            self._main_toolbar.ID,
            self.PLUGIN_NAME,
            self.CONTEXT_NAME,
        )

        self._corner_toolbar = MainWidgetToolbar(
            parent=self,
            title=_("Main widget corner toolbar"),
        )
        self._corner_toolbar.ID = "corner_toolbar"

        TOOLBAR_REGISTRY.register_reference(
            self._corner_toolbar,
            self._corner_toolbar.ID,
            self.PLUGIN_NAME,
            self.CONTEXT_NAME,
        )

        self._corner_toolbar.setSizePolicy(
            QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self._options_menu = self._create_menu(
            PluginMainWidgetMenus.Options,
            title=_("Options menu"),
            MenuClass=PluginMainWidgetOptionsMenu,
        )

        # Margins
        # --------------------------------------------------------------------
        # These margins are necessary to give some space between the widgets
        # inside this one and the window separator and borders.
        self._margin_right = AppStyle.MarginSize
        self._margin_bottom = AppStyle.MarginSize
        if not self.get_conf("vertical_tabs", section="main"):
            self._margin_left = AppStyle.MarginSize
        else:
            self._margin_left = 0

        # Layout
        # --------------------------------------------------------------------
        self._main_layout = QVBoxLayout()
        self._toolbars_layout = QVBoxLayout()
        self._main_toolbar_layout = QHBoxLayout()

        self._toolbars_layout.setContentsMargins(
            self._margin_left, 0, self._margin_right, 0
        )
        self._toolbars_layout.setSpacing(0)
        self._main_toolbar_layout.setContentsMargins(0, 0, 0, 0)
        self._main_toolbar_layout.setSpacing(0)
        self._main_layout.setContentsMargins(0, 0, 0, 0)
        self._main_layout.setSpacing(0)

        # Add inititals layouts
        self._main_toolbar_layout.addWidget(self._main_toolbar, stretch=10000)
        self._main_toolbar_layout.addWidget(self._corner_toolbar, stretch=1)
        self._toolbars_layout.addLayout(self._main_toolbar_layout)
        self._main_layout.addLayout(self._toolbars_layout, stretch=1)

        # Create a stacked layout when the widget displays an empty message
        if self.SHOW_MESSAGE_WHEN_EMPTY and self.get_conf(
            "show_message_when_panes_are_empty", section="main"
        ):
            if not self.MESSAGE_WHEN_EMPTY:
                raise SpyderAPIError(
                    "You need to provide a message to show when the widget is "
                    "empty"
                )

            self._pane_empty = EmptyMessageWidget(
                self,
                self.IMAGE_WHEN_EMPTY,
                self.MESSAGE_WHEN_EMPTY,
                self.DESCRIPTION_WHEN_EMPTY,
                adjust_on_resize=True,
            )

            self._stack = QStackedWidget(self)
            self._stack.addWidget(self._pane_empty)

            if self.SET_LAYOUT_WHEN_EMPTY:
                layout = QVBoxLayout()
                layout.addWidget(self._stack)
                self.setLayout(layout)

    # ---- Private Methods
    # -------------------------------------------------------------------------
    def _setup(self) -> None:
        """
        Setup default actions, create options menu, and connect signals.
        """
        # Tabs
        children = self.findChildren(Tabs)
        if children:
            for child in children:
                self._is_tab = True
                # For widgets that use tabs, we add the corner widget using
                # the setCornerWidget method.
                child.setCornerWidget(self._corner_widget)
                self._corner_widget.setStyleSheet(str(PANES_TABBAR_STYLESHEET))
                break

        self._options_button = self.create_toolbutton(
            PluginMainWidgetWidgets.OptionsToolButton,
            text=_("Options"),
            icon=self.create_icon("tooloptions"),
        )

        self.add_corner_widget(self._options_button)

        if self.ENABLE_SPINNER:
            self.add_corner_widget(self._spinner)

        # Widget setup
        # --------------------------------------------------------------------
        self._main_toolbar.setVisible(not self._is_tab)
        self._corner_toolbar.setVisible(not self._is_tab)
        self._options_button.setPopupMode(QToolButton.InstantPopup)

        # Create default widget actions
        self.dock_action = self.create_action(
            name=PluginMainWidgetActions.DockPane,
            text=_("Dock"),
            tip=_("Dock the pane"),
            icon=self.create_icon("dock"),
            triggered=self.dock_window,
        )
        self.lock_unlock_action = self.create_action(
            name=PluginMainWidgetActions.LockUnlockPosition,
            text=_("Move"),
            tip=_("Unlock to move pane to another position"),
            icon=self.create_icon("drag_dock_widget"),
            triggered=self.lock_unlock_position,
        )
        self.undock_action = self.create_action(
            name=PluginMainWidgetActions.UndockPane,
            text=_("Undock"),
            tip=_("Undock the pane"),
            icon=self.create_icon("undock"),
            triggered=self.create_window,
        )
        self.close_action = self.create_action(
            name=PluginMainWidgetActions.ClosePane,
            text=_("Close"),
            tip=_("Close the pane"),
            icon=self.create_icon("close_pane"),
            triggered=self.close_dock,
        )
        # We use this instead of the QDockWidget.toggleViewAction
        self.toggle_view_action = self.create_action(
            name="switch to " + self._name,
            text=self.get_title(),
            toggled=lambda checked: self.toggle_view(checked),
            context=Qt.WidgetWithChildrenShortcut,
            shortcut_context="_",
        )

        for item in [
            self.lock_unlock_action,
            self.undock_action,
            self.dock_action,
            self.close_action,
        ]:
            self.add_item_to_menu(
                item,
                self._options_menu,
                section=OptionsMenuSections.Bottom,
            )

        self._options_button.setMenu(self._options_menu)
        self._options_menu.aboutToShow.connect(self._update_actions)

        # For widgets that do not use tabs, we add the corner widget to the
        # corner toolbar
        if not self._is_tab:
            self.add_item_to_toolbar(
                self._corner_widget,
                toolbar=self._corner_toolbar,
                section="corner",
            )
            self._corner_widget.setStyleSheet(str(PANES_TOOLBAR_STYLESHEET))

        # Update title
        self.setWindowTitle(self.get_title())

    def _update_actions(self) -> None:
        """
        Refresh Options menu.
        """
        show_dock_actions = self.windowwidget is None
        self.undock_action.setVisible(show_dock_actions)
        self.lock_unlock_action.setVisible(show_dock_actions)
        self.dock_action.setVisible(not show_dock_actions)

        # Widget setup
        self.update_actions()

    @Slot(bool)
    def _on_top_level_change(self, top_level: bool) -> None:
        """
        Actions to perform when a plugin is undocked to be moved.
        """
        self.undock_action.setDisabled(top_level)

        # Change the cursor shape when dragging
        if top_level:
            QApplication.setOverrideCursor(Qt.ClosedHandCursor)
        else:
            QApplication.restoreOverrideCursor()

    @Slot(bool)
    def _on_title_bar_shown(self, visible: bool) -> None:
        """
        Actions to perform when the title bar is shown/hidden.
        """
        if visible:
            self.lock_unlock_action.setText(_("Lock"))
            self.lock_unlock_action.setIcon(self.create_icon("lock_open"))
            for method_name in ["setToolTip", "setStatusTip"]:
                method = getattr(self.lock_unlock_action, method_name)
                method(_("Lock pane to the current position"))
        else:
            self.lock_unlock_action.setText(_("Move"))
            self.lock_unlock_action.setIcon(
                self.create_icon("drag_dock_widget")
            )
            for method_name in ["setToolTip", "setStatusTip"]:
                method = getattr(self.lock_unlock_action, method_name)
                method(_("Unlock to move pane to another position"))

    # ---- Public Qt overridden methods
    # -------------------------------------------------------------------------
    def setLayout(self, layout: QLayout) -> None:
        """
        Set layout for the widget.
        """
        self._main_layout.addLayout(layout, stretch=1000000)
        super().setLayout(self._main_layout)
        layout.setContentsMargins(
            self._margin_left,
            self.MARGIN_TOP,
            self._margin_right,
            self._margin_bottom,
        )
        layout.setSpacing(0)

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Handle closing the widget.

        Parameters
        ----------
        event : QCloseEvent
            The event object closing this widget.

        Returns
        -------
        None
        """
        self.on_close()
        super().closeEvent(event)

    def focusInEvent(self, event: QFocusEvent) -> None:
        """
        Handle the widget gaining focus.

        Parameters
        ----------
        event : QFocusEvent
            The focus event object.

        Returns
        -------
        None
        """
        self.sig_focus_status_changed.emit(True)
        self.on_focus_in()
        return super().focusInEvent(event)

    def focusOutEvent(self, event: QFocusEvent) -> None:
        """
        Handle the widget losing focus.

        Parameters
        ----------
        event : QFocusEvent
            The focus event object.

        Returns
        -------
        None
        """
        self.sig_focus_status_changed.emit(False)
        self.on_focus_out()
        return super().focusOutEvent(event)

    # ---- Public methods to use
    # -------------------------------------------------------------------------
    def get_plugin(self) -> SpyderPluginV2:
        """
        Return the parent plugin of this widget.

        Returns
        -------
        SpyderPluginV2
            The parent plugin of this widget.
        """
        return self._plugin

    def get_action(
        self, name: str, context: str | None = None, plugin: str | None = None
    ) -> spyder.utils.qthelpers.SpyderAction:
        """
        Return an action by name, context and plugin.

        Parameters
        ----------
        name: str
            Identifier of the action to retrieve.
        context: str | None, optional
            Context identifier under which the action was stored.
            If ``None``, the default, then the
            :attr:`CONTEXT_NAME` attribute is used instead.
        plugin: str | None, optional
            Identifier of the plugin in which the action was defined.
            If ``None``, the default, then the
            :attr:`~spyder.api.widgets.mixins.SpyderWidgetMixin.PLUGIN_NAME`
            attribute is used instead.

        Returns
        -------
        spyder.utils.qthelpers.SpyderAction
            The corresponding action stored under the given ``name``,
             ``context`` and ``plugin``.

        Raises
        ------
        KeyError
            If the combination of ``name``, ``context`` and ``plugin`` keys
            does not exist in the action registry.
        """
        plugin = self.PLUGIN_NAME if plugin is None else plugin
        context = self.CONTEXT_NAME if context is None else context

        return ACTION_REGISTRY.get_reference(name, plugin, context)

    def add_corner_widget(
        self,
        action_or_widget: spyder.utils.qthelpers.SpyderAction | QWidget,
        before: spyder.utils.qthelpers.SpyderAction | QWidget | None = None,
    ) -> None:
        """
        Add a widget to the corner toolbar.

        By default, widgets are added to the left of the last toolbar item.
        Corner widgets provide an options menu button and a spinner so any
        additional widgets will be placed the left of the spinner, if visible
        (unless ``before`` is set).

        Parameters
        ----------
        widget : spyder.utils.qthelpers.SpyderAction | QWidget
            The action or widget to add to the toolbar.
        before : spyder.utils.qthelpers.SpyderAction | QWidget | None, optional
            The action or widget to add ``widget`` before (to the right of).
            If ``None`` (the default), the widget will be added to the left
            of the left-most widget.

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If either ``widget`` or ``before`` lacks a ``name`` attribute;
            a widget with the same ``name`` as ``widget`` was already added;
            a widget with ``before.name`` has not been added previously; or
            the first widget added is not the options (hamburger) menu widget.
        """
        self._corner_widget.add_widget(action_or_widget, before=before)

    def get_corner_widget(
        self, name: str
    ) -> spyder.utils.qthelpers.SpyderAction | QWidget | None:
        """
        Return a widget by its unique ID (i.e. its ``name`` attribute).

        Parameters
        ----------
        name : str
            The ``name`` attribute of the widget to return.

        Returns
        -------
        QWidget | None
            The widget object corresponding to ``name``, or ``None``
            if a widget with that ``name`` does not exist.
        """
        return self._corner_widget.get_widget(name)

    def start_spinner(self) -> None:
        """
        Start the default status spinner.

        Returns
        -------
        None
        """
        if self.ENABLE_SPINNER:
            self._spinner.start()

    def stop_spinner(self) -> None:
        """
        Stop the default status spinner.

        Returns
        -------
        None
        """
        if self.ENABLE_SPINNER:
            self._spinner.stop()

    def create_toolbar(self, toolbar_id: str) -> MainWidgetToolbar:
        """
        Create and add an auxiliary toolbar to the top of the plugin.

        Parameters
        ----------
        toolbar_id: str
            The unique identifier name of this toolbar.

        Returns
        -------
        MainWidgetToolbar
            The auxiliary toolbar object that was created.
        """
        toolbar = MainWidgetToolbar(parent=self)
        toolbar.ID = toolbar_id

        TOOLBAR_REGISTRY.register_reference(
            toolbar, toolbar_id, self.PLUGIN_NAME, self.CONTEXT_NAME
        )

        self._auxiliary_toolbars[toolbar_id] = toolbar
        self._toolbars_layout.addWidget(toolbar)

        return toolbar

    def get_options_menu(self) -> PluginMainWidgetOptionsMenu:
        """
        Return the options ("hamburger") menu for this widget.

        Returns
        -------
        PluginMainWidgetOptionsMenu
            The options ("hamburger") menu widget.
        """
        return self._options_menu

    def get_options_menu_button(self) -> QToolButton:
        """
        Return the options menu button for this widget.

        Returns
        -------
        QToolButton
            The button widget for the plugin options ("hamburger") menu.
        """
        return self._options_button

    def get_main_toolbar(self) -> MainWidgetToolbar:
        """
        Return the main toolbar of this widget.

        Returns
        -------
        MainWidgetToolbar
            The main toolbar of the widget that contains the options button.
        """
        return self._main_toolbar

    def get_auxiliary_toolbars(self) -> OrderedDict[MainWidgetToolbar]:
        """
        Return the auxiliary toolbars of this widget.

        Returns
        -------
        OrderedDict[MainWidgetToolbar]
            A dictionary with toolbar IDs as keys and their corresponding
            auxiliary toolbar widgets as values.
        """
        return self._auxiliary_toolbars

    def set_icon_size(self, icon_size: int) -> None:
        """
        Set the icon size of this widget's toolbars.

        Parameters
        ----------
        iconsize: int
            An integer corresponding to the size in pixels to which the icons
            of the plugin's toolbars need to be set.
        """
        self._icon_size = icon_size
        self._main_toolbar.set_icon_size(QSize(icon_size, icon_size))

    def show_status_message(self, message: str, timeout: int) -> None:
        """
        Show a message in the Spyder status bar.

        Parameters
        ----------
        message: str
            The message to display in the status bar.
        timeout: int
            The amount of time, in milliseconds, to display the message.
            If ``0``, the default, the message will be shown until a plugin
            calls :meth:`!show_status_message` again.

        Returns
        -------
        None
        """
        status_bar = self.statusBar()
        if status_bar.isVisible():
            status_bar.showMessage(message, timeout)

    def get_focus_widget(self) -> PluginMainWidget:
        """
        Get the widget to give focus to.

        This is called when this widget's dockwidget is raised to the top.

        Returns
        -------
        QWidget
            The widget to give focus to.
        """
        return self

    def update_margins(self, margin=None):
        """
        Update the margins of this widget's central widget.

        Parameters
        ----------
        margin: int | None
            The margins to use for the central widget, or ``None`` for the
            default margins.

        Returns
        -------
        None
        """
        layout = self.layout()
        if self._default_margins is None:
            self._default_margins = layout.getContentsMargins()

        if margin is not None:
            layout.setContentsMargins(margin, margin, margin, margin)
        else:
            layout.setContentsMargins(*self._default_margins)

    def update_title(self) -> None:
        """
        Update this widget's dockwidget title.

        Returns
        -------
        None
        """
        if self.dockwidget is not None:
            widget = self.dockwidget
        elif self.windowwidget is not None:
            widget = self.undocked_window
        else:
            return

        widget.setWindowTitle(self.get_title())

    def set_name(self, name: str) -> None:
        """
        Set this widget's name.

        .. note::

            Normally, this is set to the same as the plugin's name,
            :attr:`SpyderPluginV2.NAME <spyder.api.plugins.SpyderPluginV2.NAME>`.

        Parameters
        ----------
        name : str
            The name to set.

        Returns
        -------
        None
        """
        self._name = name

    def get_name(self) -> str:
        """
        Return this widget's name.

        By default, the same as the plugin's name,
        :attr:`SpyderPluginV2.NAME <spyder.api.plugins.SpyderPluginV2.NAME>`.

        Returns
        -------
        str
            The name of the widget, and normally the plugin as well.
        """
        return self._name

    def set_icon(self, icon: QIcon) -> None:
        """
        Set this widget's icon.

        Parameters
        ----------
        icon : QIcon
            The icon object to set as the widget's icon.

        Returns
        -------
        None
        """
        self._icon = icon

    def get_icon(self) -> QIcon:
        """
        Get this widget's icon.

        Returns
        -------
        QIcon
            The widget's icon object.
        """
        return self._icon

    def render_toolbars(self) -> None:
        """
        Render all toolbars of this widget.

        .. caution::

            This action can only be performed once.

        Returns
        -------
        None
        """
        # if not self._toolbars_already_rendered:
        self._main_toolbar.render()
        self._corner_toolbar.render()
        for __, toolbar in self._auxiliary_toolbars.items():
            toolbar.render()

            # self._toolbars_already_rendered = True

    # ---- For widgets with an empty message
    # -------------------------------------------------------------------------
    def set_content_widget(
        self, widget: QWidget, add_to_stack: bool = True
    ) -> None:
        """
        When there is an empty message, set the widget for actual content,

        Parameters
        ----------
        widget: QWidget
            Widget to set as the widget with actual (non-empty) content.
        add_to_stack: bool, optional
            Whether to add this widget to stacked widget that holds the empty
            message.

        Returns
        -------
        None
        """
        self._content_widget = widget

        if self._stack is not None:
            if add_to_stack:
                self._stack.addWidget(self._content_widget)
        else:
            # This is necessary to automatically set a layout for Find or the
            # Profiler when the user disables empty messages in Preferences.
            if self.SET_LAYOUT_WHEN_EMPTY:
                layout = QVBoxLayout()
                layout.addWidget(self._content_widget)
                self.setLayout(layout)

    def show_content_widget(self) -> None:
        """
        Show the widget that displays actual content instead of the empty one.

        Returns
        -------
        None
        """
        if (
            self._stack is not None
            and self._content_widget is not None
            and self._stack.indexOf(self._content_widget) != -1
        ):
            self._stack.setCurrentWidget(self._content_widget)

    def show_empty_message(self) -> None:
        """
        Show the empty message widget.

        Returns
        -------
        None
        """
        if self.SHOW_MESSAGE_WHEN_EMPTY and self.get_conf(
            "show_message_when_panes_are_empty", section="main"
        ):
            self._stack.setCurrentWidget(self._pane_empty)

    # ---- SpyderWindowWidget handling
    # -------------------------------------------------------------------------
    @Slot()
    def create_window(self) -> None:
        """
        Create an undocked window containing this widget.

        Returns
        -------
        None
        """
        logger.debug(f"Undocking plugin {self._name}")

        # Widgets
        self.windowwidget = window = SpyderWindowWidget(self)

        # If the close corner button is used
        self.windowwidget.sig_closed.connect(self.close_window)

        # Widget setup
        window.setAttribute(Qt.WA_DeleteOnClose)
        window.setCentralWidget(self)
        window.setWindowIcon(self.get_icon())
        window.setWindowTitle(self.get_title())
        window.resize(self.size())

        # Restore window geometry
        geometry = self.get_conf("window_geometry", default="")
        if geometry:
            try:
                window.restoreGeometry(
                    QByteArray().fromHex(str(geometry).encode("utf-8"))
                )

                # Move to the primary screen if the window is not placed in a
                # visible location.
                window.move_to_primary_screen()
            except Exception:
                pass

        # Dock widget setup
        if self.dockwidget:
            self.dockwidget.setFloating(False)
            self.dockwidget.setVisible(False)

        self.set_ancestor(window)
        self._update_actions()
        window.show()

    @Slot()
    def dock_window(self) -> None:
        """
        Dock an undocked window with this widget back to the main window.

        Returns
        -------
        None
        """
        logger.debug(f"Docking window of plugin {self._name}")

        # Reset undocked state
        self.set_conf("window_was_undocked_before_hiding", False)

        # This avoids trying to close the window twice: once when calling
        # _close_window below and the other when Qt calls the closeEvent of
        # windowwidget
        self.windowwidget.blockSignals(True)

        # Close window
        self._close_window(switch_to_plugin=True)

        # Make plugin visible on main window
        self.dockwidget.setVisible(True)
        self.dockwidget.raise_()

    @Slot()
    def close_window(self) -> None:
        """
        Close undocked window when clicking on the close window button.

        This can either dock or hide the window, depending on whether the
        user hid the window before:

        * The default behavior is to dock the window, so that new users can
          experiment with the dock/undock functionality without surprises.
        * If the user closes the window by clicking on the :guilabel:`Close`
          action in the widget's options ("hamburger") menu or by
          going to the :menuselection:`Window --> Panes` menu,
          then we will hide it when they click on the close button again.
          That gives users the ability to show/hide panes without
          docking/undocking them first.

        Returns
        -------
        None
        """
        if self.get_conf("window_was_undocked_before_hiding", default=False):
            self.close_dock()
        else:
            self.dock_window()

    def _close_window(
        self, save_undocked: bool = False, switch_to_plugin: bool = True
    ) -> None:
        """
        Helper function to close the undocked window with different parameters.

        Parameters
        ----------
        save_undocked : bool, optional
            ``True`` if the window state (size and position) should be saved.
            If ``False``, the default, don't persist the window state.
        switch_to_plugin : bool, optional
            Whether to switch to the plugin after closing the window.
            If ``True`` (the default), will switch to the plugin.

        Returns
        -------
        None
        """
        if self.windowwidget is not None:
            # Save window geometry to restore it when undocking the plugin
            # again.
            geometry = self.windowwidget.saveGeometry()
            self.set_conf("window_geometry", qbytearray_to_str(geometry))

            # Save undocking state if requested
            if save_undocked:
                self.set_conf("undocked_on_window_close", True)

            # Fixes spyder-ide/spyder#10704
            self.__unsafe_window = self.windowwidget
            self.__unsafe_window.deleteLater()
            self.windowwidget.close()
            self.windowwidget = None

            # These actions can appear disabled when 'Dock' action is pressed
            self.undock_action.setDisabled(False)
            self.close_action.setDisabled(False)

            if self.dockwidget is not None:
                self.sig_update_ancestor_requested.emit()
                if switch_to_plugin:
                    # This is necessary to restore the main window layout when
                    # there's a maximized plugin on it when the user requests
                    # to dock back this plugin.
                    self.get_plugin().switch_to_plugin()

                self.dockwidget.setWidget(self)
                self._update_actions()
        else:
            # Reset undocked state
            self.set_conf("undocked_on_window_close", False)

    # ---- SpyderDockwidget handling
    # -------------------------------------------------------------------------
    def change_visibility(
        self, enable: bool, force_focus: bool | None = None
    ) -> None:
        """
        Raise this widget to the foreground, and/or grab its focus.

        Parameters
        ----------
        state : bool
            Whether the widget is being raised to the foreground
            (``True``) or set as not in the foreground (``False``).
            The latter does not actually send it to the background, but
            does configure it for not being actively shown (e.g. it disables
            its empty pane widget, if any).
        force_focus : bool | None, optional
            If ``True``, always give the widget keyboard focus when
            raising or un-raising it with this method. If ``None``, only give
            it focus when showing, not hiding (setting ``state`` to ``True``),
            and only if
            :attr:`SpyderDockablePlugin.RAISE_AND_FOCUS <spyder.api.plugins.SpyderDockablePlugin.RAISE_AND_FOCUS>`
            is ``True``. If ``False``, the default, don't give it focus
            regardless.

        Returns
        -------
        None
        """
        if self.dockwidget is None:
            return

        if enable:
            # Avoid double trigger of visibility change
            self.dockwidget.blockSignals(True)
            self.dockwidget.raise_()
            self.dockwidget.blockSignals(False)

        raise_and_focus = getattr(self, "RAISE_AND_FOCUS", None)

        if force_focus is None:
            if raise_and_focus and enable:
                focus_widget = self.get_focus_widget()
                if focus_widget:
                    focus_widget.setFocus()
        elif force_focus is True:
            focus_widget = self.get_focus_widget()
            if focus_widget:
                focus_widget.setFocus()
        elif force_focus is False:
            pass

        # If the widget is undocked, it's always visible
        self.is_visible = enable or (self.windowwidget is not None)

        if (
            self.SHOW_MESSAGE_WHEN_EMPTY
            and self.get_conf(
                "show_message_when_panes_are_empty", section="main"
            )
            # We need to do this validation to prevent errors after changing
            # the option above in Preferences and restarting Spyder.
            and self._pane_empty is not None
        ):
            self._pane_empty.set_visibility(self.is_visible)

        # TODO: Pending on plugin migration that uses this
        # if getattr(self, 'DISABLE_ACTIONS_WHEN_HIDDEN', None):
        #     for __, action in self.get_actions().items():
        #         action.setEnabled(is_visible)

    def toggle_view(self, checked: bool) -> None:
        """
        Show or hide this widget in the Spyder interface.

        Used to show or hide it from the from the
        :menuselection:`Window --> Panes` menu.

        Parameters
        ----------
        value : bool
            Whether to show (``True``) or hide (``False``) this widget.

        Returns
        -------
        None

        Notes
        -----
        If you need to attach some functionality when this changes, use
        :attr:`sig_toggle_view_changed`. For an example, please see
        :mod:`spyder.plugins.onlinehelp.widgets`.
        """
        if not self.dockwidget:
            return

        # To check if the plugin needs to be undocked at the end
        undock = False

        if checked:
            self.dockwidget.show()
            self.dockwidget.raise_()
            self.is_visible = True

            # We need to undock the plugin if that was its state before
            # toggling its visibility.
            if (
                # Don't run this while the window is being created to not
                # affect setting up the layout at startup.
                not self._plugin.main.is_setting_up
                and self.get_conf(
                    "window_was_undocked_before_hiding", default=False
                )
            ):
                undock = True
        else:
            if self.windowwidget is not None:
                logger.debug(f"Closing window of plugin {self._name}")

                # This avoids trying to close the window twice: once when
                # calling _close_window below and the other when Qt calls the
                # closeEvent of windowwidget
                self.windowwidget.blockSignals(True)

                # Dock plugin if it's undocked before hiding it.
                self._close_window(switch_to_plugin=False)

                # Save undocked state to restore it afterwards.
                self.set_conf("window_was_undocked_before_hiding", True)

            self.dockwidget.hide()
            self.is_visible = False

        # Update toggle view status, if needed, without emitting signals.
        if self.toggle_view_action.isChecked() != checked:
            self.blockSignals(True)
            self.toggle_view_action.setChecked(checked)
            self.blockSignals(False)

        self.sig_toggle_view_changed.emit(checked)

        logger.debug(
            f"Plugin {self._name} is now {'visible' if checked else 'hidden'}"
        )

        if undock:
            # We undock the plugin at this point so that the Window menu is
            # updated correctly.
            self.create_window()

    def create_dockwidget(
        self, mainwindow: spyder.app.mainwindow.MainWindow
    ) -> tuple[spyder.widgets.dock.SpyderDockWidget, Qt.DockWidgetArea]:
        """
        Add this widget to the parent Spyder main window as a dock widget.

        Parameters
        ----------
        mainwindow : spyder.app.mainwindow.MainWindow
            The main window to set as the dockwidget's parent.

        Returns
        -------
        spyder.widgets.dock.SpyderDockWidget
            The newly created dock widget.
        Qt.DockWidgetArea
            The area of the window the dockwidget is placed in.
        """
        # Creating dock widget
        title = self.get_title()
        self.dockwidget = dock = SpyderDockWidget(title, mainwindow)

        # Setup
        dock.setObjectName(self.__class__.__name__ + "_dw")
        dock.setWidget(self)

        # Signals
        dock.visibilityChanged.connect(self.change_visibility)
        dock.topLevelChanged.connect(self._on_top_level_change)
        dock.sig_title_bar_shown.connect(self._on_title_bar_shown)

        return (dock, dock.LOCATION)

    @Slot()
    def close_dock(self) -> None:
        """
        Close the dockwidget.

        Returns
        -------
        None
        """
        logger.debug(f"Hiding plugin {self._name}")
        self.toggle_view_action.setChecked(False)

    def lock_unlock_position(self) -> None:
        """
        Show/hide title bar to move/lock this widget's position.

        Returns
        -------
        None
        """
        if isinstance(self.dockwidget.titleBarWidget(), DockTitleBar):
            self.dockwidget.remove_title_bar()
        else:
            self.dockwidget.set_title_bar()

    def get_maximized_state(self) -> bool:
        """
        Get this widget's maximized state.

        Returns
        -------
        bool
            ``True`` if the widget is maximized, ``False`` otherwise.
        """
        return self._is_maximized

    def set_maximized_state(self, state: bool) -> None:
        """
        Set the attribute that holds this widget's maximized state.

        Parameters
        ----------
        state: bool
            ``True`` to set the widget as maximized, ``False`` set it as not
            maximized.

        Returns
        -------
        None
        """
        self._is_maximized = state

    # ---- API: methods to define or override
    # ------------------------------------------------------------------------
    def get_title(self) -> str:
        """
        Return the title that will be displayed on dockwidgets or windows.

        Returns
        -------
        str
            This dockwidget's tab/window title.

        Raises
        ------
        NotImplementedError
            If the main widget subclass doesn't define a ``get_title`` method.
        """
        raise NotImplementedError("PluginMainWidget must define `get_title`!")

    def set_ancestor(self, ancestor: QWidget) -> None:
        """
        Update the ancestor/parent of child widgets when undocking.

        Parameters
        ----------
        ancestor: QWidget
            The window widget to set as a parent of this one.

        Returns
        -------
        None
        """
        pass

    def setup(self) -> None:
        """
        Create widget actions, add to menus and perform other setup steps.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            If the main widget subclass doesn't define a ``setup`` method.
        """
        raise NotImplementedError(
            f"{type(self)} must define a `setup` method!"
        )

    def update_actions(self) -> None:
        """
        Update the state of exposed actions.

        Exposed actions are actions created by the
        :meth:`~spyder.api.widgets.mixins.SpyderActionMixin.create_action`
        method.

        Returns
        -------
        None

        Raises
        ------
        NotImplementedError
            If the subclass doesn't define an ``update_actions`` method.
        """
        raise NotImplementedError(
            "A PluginMainWidget subclass must define an `update_actions` "
            f"method! Hint: {type(self)} should implement `update_actions`"
        )

    def on_close(self) -> None:
        """
        Perform actions before the widget is closed.

        Does nothing by default; intended to be overridden for widgets
        that need to perform actions on close.

        .. warning::

            This method **must** only operate on local attributes.

        Returns
        -------
        None
        """
        pass

    def on_focus_in(self) -> None:
        """
        Perform actions when the widget receives focus.

        Does nothing by default; intended to be overridden for widgets
        that need to perform actions on gaining focus.

        Returns
        -------
        None
        """
        pass

    def on_focus_out(self) -> None:
        """
        Perform actions when the widget loses focus.

        Does nothing by default; intended to be overridden for widgets
        that need to perform actions on loosing focus.

        Returns
        -------
        None
        """
        pass


def _run_test() -> None:
    # Third party imports
    from qtpy.QtWidgets import QHBoxLayout, QTableWidget, QMainWindow

    # Local imports
    from spyder.utils.qthelpers import qapplication

    app = qapplication()
    main = QMainWindow()
    widget = PluginMainWidget("test", main)
    widget.get_title = lambda x=None: "Test title"
    widget._setup()
    layout = QHBoxLayout()
    layout.addWidget(QTableWidget())
    widget.setLayout(layout)
    widget.start_spinner()
    dock, location = widget.create_dockwidget(main)
    main.addDockWidget(location, dock)
    main.setStyleSheet(str(APP_STYLESHEET))
    main.show()
    app.exec_()


if __name__ == "__main__":
    _run_test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

from __future__ import annotations

import asyncio
import logging
import typing
from abc import abstractmethod

import aiohttp
import yarl

from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.api.utils import ABCMeta, abstract_attribute
from spyder.plugins.remoteclient import SPYDER_PLUGIN_NAME

if typing.TYPE_CHECKING:
    from spyder.plugins.remoteclient.api.manager.base import (
        SpyderRemoteAPIManagerBase,
    )


SpyderBaseJupyterAPIType = typing.TypeVar(
    "SpyderBaseJupyterAPIType", bound="SpyderBaseJupyterAPI"
)


logger = logging.getLogger(__name__)

REQUEST_TIMEOUT = 60  # seconds
VERIFY_SSL = True


class SpyderRemoteAPIError(Exception):
    """
    Exception for errors related to Spyder remote client API.
    """
    ...


class SpyderRemoteSessionClosed(SpyderRemoteAPIError):
    """
    Exception for errors related to a closed session.
    """
    ...


class SpyderRemoteConnectionError(SpyderRemoteAPIError):
    """Exception for errors related to connection establishment."""


class SpyderBaseJupyterAPI(metaclass=ABCMeta):
    """
    Base class for Jupyter API plugins.

    This class must be subclassed to implement the API for a specific
    Jupyter extension. It provides a connection method and context manager.
    """

    request_timeout = REQUEST_TIMEOUT
    verify_ssl = VERIFY_SSL

    @abstract_attribute
    def base_url(self) -> str: ...

    @property
    def server_id(self):
        """Server ID or configuration ID of the remote server."""
        return self.manager.config_id

    @property
    def server_name(self):
        """Server name of the remote server."""
        return self.manager.server_name

    def __init__(self, manager: SpyderRemoteAPIManagerBase):
        self.manager = manager
        self._session: typing.Optional[aiohttp.ClientSession] = None

    @property
    def session(self) -> aiohttp.ClientSession:
        error_message = (
            "Session is closed, please ensure that an active session is open "
            "before calling this method.\n\n"
            "You can open a session using a context manager:\n"
            f"with {type(self).__name__}() as api:\n"
            "    ...\n\n"
            "Or by calling the connect method:\n"
            f"api = {type(self).__name__}()\n"
            "await api.connect()"
        )
        if self._session is None or self._session.closed:
            raise SpyderRemoteSessionClosed(error_message)

        return self._session

    @session.setter
    def session(self, session: aiohttp.ClientSession):
        self._session = session

    @property
    def server_url(self) -> yarl.URL:
        return yarl.URL(self.manager.server_url)

    @property
    def api_url(self) -> yarl.URL:
        return self.server_url / self.base_url

    @property
    def closed(self):
        if self._session is None:
            return True
        return self._session.closed

    async def connect(self):
        """
        Establish connection to the remote Spyder's Jupyter server.

        Raises
        ------
        SpyderRemoteConnectionError
            If the connection to the remote server could not be established.
        """
        if not await AsyncDispatcher(
            loop="asyncssh",
            return_awaitable=True,
        )(self.manager.ensure_connection_and_server)():
            msg = (
                f"Could not connect to spyder "
                f"remote server at '{self.server_name}'"
            )
            raise SpyderRemoteConnectionError(msg)
        if not self.closed:
            return
        self.session = aiohttp.ClientSession(
            headers={"Authorization": f"token {self.manager.api_token}"},
            connector=aiohttp.TCPConnector(
                ssl=None if self.verify_ssl else False
            ),
            raise_for_status=self._raise_for_status,
            timeout=aiohttp.ClientTimeout(total=self.request_timeout),
        )

    async def __aenter__(self):
        await self.connect()
        return self

    async def close(self):
        if self._session is not None:
            await self._session.close()

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()

    @abstractmethod
    async def _raise_for_status(self, response: aiohttp.ClientResponse): ...

    @staticmethod
    def retry(
        num_retries=5,
        delay=1,
        check: typing.Callable[..., bool] = lambda x: bool(x),
        catch: typing.Union[type[Exception], None] = Exception,
    ):
        """
        Retry method decorator for async functions.

        Parameters
        ----------
        num_retries: int
            Number of retries to attempt.
        delay: int
            Delay in seconds between retries.
        check: callable
            Function to check if the return value is valid.
        exception: Exception, None. Defaults to Exception.
            If provided, catch this exception and retry.

        Returns
        -------
        wrapper: callable
            Decorated function.

        Example
        -------
        ```
        @retry(num_retries=5, delay=1, check=lambda x: x is not None)
        async def my_function():
            return await some_async_function()
        ```
        """
        if catch is None:

            async def run_try(tryth, func, *args, **kwargs):
                ret = await func(*args, **kwargs)
                if check(ret):
                    return ret
                logger.debug(
                    f"retry={tryth} delay={delay} invalid return={ret}"
                )
                await asyncio.sleep(delay)
        else:

            async def run_try(tryth, func, *args, **kwargs):
                try:
                    ret = await func(*args, **kwargs)
                except catch as err:
                    logger.debug(f"retry={tryth} delay={delay} error={err}")
                    await asyncio.sleep(delay)
                else:
                    if check(ret):
                        return ret
                    logger.debug(
                        f"retry={tryth} delay={delay} invalid return={ret}"
                    )
                    await asyncio.sleep(delay)

        def decorator(func):
            async def wrapper(*args, **kwargs):
                for tryth in range(num_retries - 1):
                    return await run_try(tryth, func, *args, **kwargs)

                return await func(*args, **kwargs)

            return wrapper

        return decorator


class JupyterAPI(SpyderBaseJupyterAPI):
    base_url = "api"

    def __init__(self, manager: SpyderRemoteAPIManagerBase, verify_ssl=True):
        """
        For JupyterAPI, the manager.server_url is expected to be the notebook
        URL.
        """
        super().__init__(manager)
        self.verify_ssl = verify_ssl

    async def _raise_for_status(self, response: aiohttp.ClientResponse):
        response.raise_for_status()

    # ---- Jupyter Server REST API
    @SpyderBaseJupyterAPI.retry()
    async def create_kernel(self, kernel_spec=None, spyder_kernel=True):
        data = {}
        if spyder_kernel:
            data["spyder_kernel"] = True

        if kernel_spec := (
            kernel_spec
            or self.manager.options.get("default_kernel_spec")
        ):
            data["name"] = kernel_spec

        async with self.session.post(
            self.api_url / "kernels", json=data
        ) as response:
            return await response.json()

    async def list_kernel_specs(self):
        async with self.session.get(self.api_url / "kernelspecs") as response:
            return await response.json()

    async def list_kernels(self):
        async with self.session.get(self.api_url / "kernels") as response:
            return await response.json()

    @SpyderBaseJupyterAPI.retry()
    async def get_kernel(self, kernel_id):
        try:
            async with self.session.get(
                self.api_url / "kernels" / kernel_id
            ) as response:
                return await response.json()
        except aiohttp.ClientResponseError as e:
            if e.status == 404:
                return {}
            else:
                raise e

    async def terminate_kernel(self, kernel_id):
        async with self.session.delete(
            self.api_url / "kernels" / kernel_id
        ) as response:
            if response.status == 204:
                logger.debug(f"deleted kernel={kernel_id} for jupyter")
                return True
            else:
                return False

    async def interrupt_kernel(self, kernel_id):
        async with self.session.post(
            self.api_url / "kernels" / kernel_id / "interrupt"
        ) as response:
            if response.status == 204:
                logger.debug(f"interrupted kernel={kernel_id} for jupyter")
                return True
            else:
                return False

    async def restart_kernel(self, kernel_id):
        async with self.session.post(
            self.api_url / "kernels" / kernel_id / "restart"
        ) as response:
            if response.status == 200:
                logger.debug(f"restarted kernel={kernel_id} for jupyter")
                return True
            else:
                return False

    async def shutdown_server(self):
        async with self.session.post(self.api_url / "shutdown") as response:
            if response.status == 200:
                logger.debug("shutdown jupyter server")
                return True
            else:
                return False

    async def get_plugin_version(self):
        """Get the version of the Jupyter server."""
        try:
            async with self.session.get(
                self.server_url / SPYDER_PLUGIN_NAME / "version",
            ) as response:
                return await response.text()
        except aiohttp.ClientError as e:
            msg = "Failed to get plugin version"
            raise SpyderRemoteAPIError(msg) from e



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Container Widget.

Holds references for base actions in the Application of Spyder.
"""

# Standard library imports
import logging

# Third party imports
from qtpy.QtCore import Slot

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.main_container import PluginMainContainer
from spyder.plugins.updatemanager.widgets.status import UpdateManagerStatus
from spyder.plugins.updatemanager.widgets.update import UpdateManagerWidget
from spyder.utils.qthelpers import DialogManager

# Logger setup
logger = logging.getLogger(__name__)


# Actions
class UpdateManagerActions:
    SpyderCheckUpdateAction = "spyder_check_update_action"


class UpdateManagerContainer(PluginMainContainer):

    def __init__(self, name, plugin, parent=None):
        super().__init__(name, plugin, parent)

        self.install_on_close = False

    # ---- PluginMainContainer API
    # -------------------------------------------------------------------------
    def setup(self):
        self.dialog_manager = DialogManager()
        self.update_manager = UpdateManagerWidget(parent=self)
        self.update_manager_status = UpdateManagerStatus(parent=self)

        # Actions
        self.check_update_action = self.create_action(
            UpdateManagerActions.SpyderCheckUpdateAction,
            _("Check for updates"),
            triggered=self.start_check_update
        )

        # Signals
        self.update_manager.sig_set_status.connect(self.set_status)
        self.update_manager.sig_disable_actions.connect(
            self._set_actions_state
        )
        self.update_manager.sig_block_status_signals.connect(
            self.update_manager_status.blockSignals)
        self.update_manager.sig_download_progress.connect(
            self.update_manager_status.set_download_progress)
        self.update_manager.sig_exception_occurred.connect(
            self.sig_exception_occurred
        )
        self.update_manager.sig_install_on_close.connect(
            self.set_install_on_close)
        self.update_manager.sig_quit_requested.connect(self.sig_quit_requested)

        self.update_manager_status.sig_check_update.connect(
            self.start_check_update)
        self.update_manager_status.sig_start_update.connect(self.start_update)
        self.update_manager_status.sig_show_progress_dialog.connect(
            self.update_manager.show_progress_dialog)

    def update_actions(self):
        pass

    def on_close(self):
        """To call from Spyder when the plugin is closed."""
        self.update_manager.cleanup_threads()

        # Run installer after Spyder is closed
        if self.install_on_close:
            self.update_manager.start_install()

        self.dialog_manager.close_all()

    # ---- Public API
    # -------------------------------------------------------------------------
    def set_status(self, status, latest_version=None):
        """Set Update Manager status"""
        self.update_manager_status.set_value(status)

    @Slot()
    def start_check_update(self, startup=False):
        """Check for spyder updates."""
        self.update_manager.start_check_update(startup=startup)

    @Slot()
    def start_update(self):
        """Start the update process"""
        self.update_manager.start_update()

    def set_install_on_close(self, install_on_close):
        """Set whether start install on close."""
        self.install_on_close = install_on_close

    # ---- Private API
    # -------------------------------------------------------------------------
    @Slot(bool)
    def _set_actions_state(self, is_disabled):
        self.check_update_action.setDisabled(is_disabled)

        # Change text to give better feedback to users about why the action is
        # disabled.
        if is_disabled:
            self.check_update_action.setText(_("Checking for updates..."))
        else:
            self.check_update_action.setText(_("Check for updates"))



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Preferences Plugin API.
"""


class PreferencesActions:
    Show = 'show_action'
    Reset = 'reset_action'



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Spyder Switcher API.
"""

class SwitcherActions:
    FileSwitcherAction = 'file switcher'
    SymbolFinderAction = 'symbol finder'



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Help Plugin Configuration Page."""

# Third party imports
from qtpy.QtWidgets import QGroupBox, QVBoxLayout

# Local imports
from spyder.api.preferences import PluginConfigPage
from spyder.api.translations import _
from spyder.utils.icon_manager import ima


class HistoryConfigPage(PluginConfigPage):
    """Configuration page for History plugin."""

    def get_icon(self):
        """Get icon to use in Configurations dialog."""
        return ima.icon('history')

    def setup_page(self):
        """Setup config page widgets and options."""
        sourcecode_group = QGroupBox(_("Display"))
        wrap_mode_box = self.create_checkbox(_("Wrap lines"), 'wrap')
        linenumbers_mode_box = self.create_checkbox(_("Show line numbers"),
                                                    'line_numbers')
        go_to_eof_box = self.create_checkbox(
                        _("Scroll automatically to last entry"), 'go_to_eof')

        sourcecode_layout = QVBoxLayout()
        sourcecode_layout.addWidget(wrap_mode_box)
        sourcecode_layout.addWidget(linenumbers_mode_box)
        sourcecode_layout.addWidget(go_to_eof_box)
        sourcecode_group.setLayout(sourcecode_layout)

        vlayout = QVBoxLayout()
        vlayout.addWidget(sourcecode_group)
        vlayout.addStretch(1)
        self.setLayout(vlayout)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)
# -----------------------------------------------------------------------------

"""Kernel Client subclass."""

# Standard library imports
import socket

# Third party imports
import asyncssh
from qtpy.QtCore import Signal
from qtconsole.client import QtKernelClient, QtZMQSocketChannel
from traitlets import Type

# Local imports
from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.api.translations import _
from spyder.plugins.ipythonconsole import SpyderKernelError


class KernelClientTunneler:
    """Class to handle SSH tunneling for a kernel connection."""

    def __init__(self, ssh_connection, *, _close_conn_on_exit=False):
        self.ssh_connection = ssh_connection
        self._port_forwarded = {}
        self._close_conn_on_exit = _close_conn_on_exit

    def __del__(self):
        """Close all port forwarders and the connection if required."""
        for forwarder in self._port_forwarded.values():
            forwarder.close()

        if self._close_conn_on_exit:
            self.ssh_connection.close()

    @classmethod
    @AsyncDispatcher(loop="asyncssh", early_return=False)
    async def new_connection(cls, *args, **kwargs):
        """Create a new SSH connection."""
        return cls(
            await asyncssh.connect(*args, **kwargs, known_hosts=None),
            _close_conn_on_exit=True,
        )

    @classmethod
    def from_connection(cls, ssh_connection):
        """Create a new KernelTunnelHandler from an existing connection."""
        return cls(ssh_connection)

    @AsyncDispatcher(loop="asyncssh", early_return=False)
    async def forward_port(self, remote_host, remote_port):
        """Forward a port through the SSH connection."""
        local = self._get_free_port()
        try:
            self._port_forwarded[(remote_host, remote_port)] = (
                await self.ssh_connection.forward_local_port(
                    '127.0.0.1', local, remote_host, remote_port
                )
            )
        except asyncssh.Error as err:
            raise SpyderKernelError(
                _(
                    "It was not possible to open an SSH tunnel to connect to "
                    "the remote kernel. Please check your credentials and the "
                    "server connection status."
                )
            ) from err
        return local

    @staticmethod
    def _get_free_port():
        """Request a free port from the OS."""
        with socket.socket() as s:
            s.bind(("", 0))
            return s.getsockname()[1]


class SpyderKernelClient(QtKernelClient):
    # Enable receiving messages on control channel.
    # Useful for pdb completion
    control_channel_class = Type(QtZMQSocketChannel)
    sig_spyder_kernel_info = Signal(object)

    def _handle_kernel_info_reply(self, rep):
        """Check spyder-kernels version."""
        super()._handle_kernel_info_reply(rep)
        spyder_kernels_info = rep["content"].get("spyder_kernels_info", None)
        self.sig_spyder_kernel_info.emit(spyder_kernels_info)

    def tunnel_to_kernel(
        self, hostname=None, sshkey=None, password=None, ssh_connection=None
    ):
        """Tunnel to remote kernel."""
        if ssh_connection is not None:
            self.__tunnel_handler = KernelClientTunneler.from_connection(
                ssh_connection
            )
        elif sshkey is not None:
            self.__tunnel_handler = KernelClientTunneler.new_connection(
                password=password,
                client_keys=[sshkey],
                **self._split_shh_address(hostname),
            )
        else:
            self.__tunnel_handler = KernelClientTunneler.new_connection(
                password=password,
                **self._split_shh_address(hostname),
            )

        (
            self.shell_port,
            self.iopub_port,
            self.stdin_port,
            self.hb_port,
            self.control_port,
        ) = (
            self.__tunnel_handler.forward_port(self.ip, port)
            for port in (
                self.shell_port,
                self.iopub_port,
                self.stdin_port,
                self.hb_port,
                self.control_port,
            )
        )
        self.ip = "127.0.0.1"  # Tunneled to localhost

    @staticmethod
    def _split_shh_address(address):
        """Split ssh address into host and port."""
        user_host, _, port = address.partition(':')
        user, _, host = user_host.rpartition('@')
        return {
            'username': user if user else None,
            'host': host if host else None,
            'port': int(port) if port else None
        }



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Plots Plugin.
"""

# Local imports
from spyder.api.plugins import Plugins, SpyderDockablePlugin
from spyder.api.shellconnect.mixins import ShellConnectPluginMixin
from spyder.api.translations import _
from spyder.plugins.plots.widgets.main_widget import PlotsWidget


class Plots(SpyderDockablePlugin, ShellConnectPluginMixin):
    """
    Plots plugin.
    """
    NAME = 'plots'
    REQUIRES = [Plugins.IPythonConsole]
    TABIFY = [Plugins.VariableExplorer, Plugins.Help]
    WIDGET_CLASS = PlotsWidget
    CONF_SECTION = NAME
    CONF_FILE = False
    DISABLE_ACTIONS_WHEN_HIDDEN = False

    # ---- SpyderDockablePlugin API
    # ------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _('Plots')

    @staticmethod
    def get_description():
        return _('View, browse and save generated figures.')

    @classmethod
    def get_icon(cls):
        return cls.create_icon('plot')

    def on_initialize(self):
        # If a figure is loaded, raise the dockwidget the first time
        # a plot is generated.
        self.get_widget().sig_figure_loaded.connect(self._on_first_plot)

    # ---- Public API
    # ------------------------------------------------------------------------
    def add_plot(self, fig, fmt, shellwidget):
        """
        Add a plot to the specified figure browser.

        Add the plot to the figure browser with the given shellwidget. Also,
        bring the plugin to the front and raise the window that it is in so
        that the plot is shown.

        If no figure browser with the given shellwidget exists, then nothing
        happens.
        """
        self.switch_to_plugin(force_focus=False)
        self.get_widget().window().raise_()
        self.get_widget().add_plot(fig, fmt, shellwidget)

    # ---- Private API
    # ------------------------------------------------------------------------
    def _on_first_plot(self):
        """Actions to execute after the first plot is generated."""
        # Only switch when inline plotting is muted. This avoids
        # showing the plugin when users want to only see plots in
        # the IPython console.
        # Fixes spyder-ide/spyder#15467
        if self.get_conf('mute_inline_plotting'):
            self.switch_to_plugin(force_focus=False)

        # We only give raise to the plugin once per session, to let users
        # know that plots are displayed in this plugin.
        # Fixes spyder-ide/spyder#15705
        self.get_widget().sig_figure_loaded.disconnect(self._on_first_plot)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Shortcuts utils."""

from dataclasses import dataclass
from typing import List, Optional

from qtpy.QtCore import QObject


@dataclass(frozen=True)
class ShortcutData:
    """Dataclass to represent shortcut data."""

    qobject: Optional[QObject]
    """
    QObject to which the shortcut will be associated.

    Notes
    -----
    This can be None when there's no need to register the shortcut to a
    specific QObject.
    """

    name: str
    """Shortcut name (e.g. "run cell")."""

    context: str
    """
    Name of the shortcut context.

    Notes
    -----
    This can be the plugin name (e.g. "editor" for shortcuts that have
    effect when the Editor is focused) or "_" for global shortcuts.
    """

    plugin_name: Optional[str] = None
    """
    Name of the plugin where the shortcut is defined.

    Notes
    -----
    This is only necessary for third-party plugins that have shortcuts with
    several contexts.
    """

    add_shortcut_to_tip: bool = False
    """Whether to add the shortcut to the qobject's tooltip."""


# List to save shortcut data registered for all widgets
SHORTCUTS_FOR_WIDGETS_DATA: List[ShortcutData] = []



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Widget to show a friendly message when another one is empty.

It's used in combination with a QStackedWidget.
"""

from __future__ import annotations
import textwrap

import qstylizer.style
import qtawesome as qta
from qtpy.QtCore import QRect, QSize, Qt
from qtpy.QtGui import QFontMetrics
from qtpy.QtWidgets import QFrame, QLabel, QSpacerItem, QVBoxLayout
from superqt.utils import qdebounced

from spyder.api.fonts import SpyderFontType, SpyderFontsMixin
from spyder.api.widgets.mixins import SvgToScaledPixmap
from spyder.utils.icon_manager import ima
from spyder.utils.palette import SpyderPalette


class EmptyMessageWidget(QFrame, SvgToScaledPixmap, SpyderFontsMixin):
    """Widget to show a friendly message when another one is empty."""

    def __init__(
        self,
        parent,
        icon_filename: str | None = None,
        text: str | None = None,
        description: str | None = None,
        top_stretch: int = 1,
        middle_stretch: int = 1,
        bottom_stretch: int = 0,
        spinner: bool = False,
        adjust_on_resize: bool = False,
        highlight_on_focus_in: bool = True,
    ):
        super().__init__(parent)

        # Attributes
        self._description = description
        self._adjust_on_resize = adjust_on_resize
        self._highlight_on_focus_in = highlight_on_focus_in
        self._is_shown = False
        self._spin = None
        self._min_height = None
        self._is_visible = False

        # This is public so it can be overridden in subclasses
        self.css = qstylizer.style.StyleSheet()

        # This is necessary to make Qt reduce the size of all widgets on
        # vertical resizes
        if self._adjust_on_resize:
            self.setMinimumHeight(150)

        interface_font_size = self.get_font(
            SpyderFontType.Interface).pointSize()

        # Image (icon)
        image_label_qss = qstylizer.style.StyleSheet()
        image_label_qss.QLabel.setValues(border="0px")

        self._image_label = None
        if icon_filename:
            self._image_label = QLabel(self)
            self._image_label.setPixmap(
                self.svg_to_scaled_pixmap(icon_filename, rescale=0.8)
            )
            self._image_label.setAlignment(Qt.AlignCenter)

            self._image_label.setStyleSheet(image_label_qss.toString())

        # Main text
        if text is not None:
            text_label = QLabel(text, parent=self)
            text_label.setAlignment(Qt.AlignCenter)
            text_label.setWordWrap(True)
            text_label_qss = qstylizer.style.StyleSheet()
            text_label_qss.QLabel.setValues(
                fontSize=f"{interface_font_size + 5}pt", border="0px"
            )
            text_label.setStyleSheet(text_label_qss.toString())

        # Description text
        self._description_label = None
        if self._description is not None:
            self._description_label = QLabel(self._description, parent=self)
            self._description_label.setAlignment(Qt.AlignCenter)
            self._description_label.setWordWrap(True)
            self._description_label.setScaledContents(True)

            description_label_qss = qstylizer.style.StyleSheet()
            description_label_qss.QLabel.setValues(
                fontSize=f"{interface_font_size}pt",
                backgroundColor=SpyderPalette.COLOR_OCCURRENCE_3,
                border="0px",
                padding="20px",
            )
            self._description_label.setStyleSheet(
                description_label_qss.toString()
            )

        # Setup layout
        pane_empty_layout = QVBoxLayout()

        # Add the top stretch
        pane_empty_layout.addStretch(top_stretch)

        # Add the image_label (icon)
        if icon_filename is not None:
            pane_empty_layout.addWidget(self._image_label)

        # Display spinner if requested
        if spinner:
            spin_widget = qta.IconWidget()
            self._spin = qta.Spin(spin_widget, interval=3, autostart=False)
            spin_icon = qta.icon(
                "mdi.loading",
                color=ima.MAIN_FG_COLOR,
                animation=self._spin
            )

            spin_widget.setIconSize(QSize(32, 32))
            spin_widget.setIcon(spin_icon)
            spin_widget.setStyleSheet(image_label_qss.toString())
            spin_widget.setAlignment(Qt.AlignCenter)

            pane_empty_layout.addWidget(spin_widget)
            pane_empty_layout.addItem(QSpacerItem(20, 20))

        # If text, display text and stretch
        if text is not None:
            pane_empty_layout.addWidget(text_label)
            pane_empty_layout.addStretch(middle_stretch)

        # If description, display description
        if self._description is not None:
            pane_empty_layout.addWidget(self._description_label)

        pane_empty_layout.addStretch(bottom_stretch)
        pane_empty_layout.setContentsMargins(20, 20, 20, 20)
        self.setLayout(pane_empty_layout)

        # Setup style
        self.setFocusPolicy(Qt.StrongFocus)
        self._apply_stylesheet(False)

    # ---- Public methods
    # -------------------------------------------------------------------------
    def setup(self, *args, **kwargs):
        """
        This method is needed when using this widget to show a "no connected
        console" message in plugins that inherit from ShellConnectMainWidget.
        """
        pass

    def set_visibility(self, visible):
        """Adjustments when the widget's visibility changes."""
        self._is_visible = visible

        if self._adjust_on_resize and self._image_label is not None:
            if visible:
                if (
                    self._min_height is not None
                    and self.height() >= self._min_height
                ):
                    self._image_label.show()
            else:
                self._image_label.hide()

    # ---- Qt methods
    # -------------------------------------------------------------------------
    def showEvent(self, event):
        """Adjustments when the widget is shown."""
        if not self._is_shown:
            self._start_spinner()
            self._is_shown = True

            if self._adjust_on_resize and self._min_height is None:
                self._min_height = self.minimumSizeHint().height()

        super().showEvent(event)

    def hideEvent(self, event):
        """Adjustments when the widget is hidden."""
        self._stop_spinner()
        self._is_shown = False
        super().hideEvent(event)

    def focusInEvent(self, event):
        if self._highlight_on_focus_in:
            self._apply_stylesheet(True)
        super().focusOutEvent(event)

    def focusOutEvent(self, event):
        if self._highlight_on_focus_in:
            self._apply_stylesheet(False)
        super().focusOutEvent(event)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        if self._adjust_on_resize:
            self._on_resize_event()

    # ---- Private methods
    # -------------------------------------------------------------------------
    def _apply_stylesheet(self, focus):
        if focus:
            border_color = SpyderPalette.COLOR_ACCENT_3
        else:
            border_color = SpyderPalette.COLOR_BACKGROUND_4

        self.css.QFrame.setValues(
            border=f'1px solid {border_color}',
            margin='0px',
            padding='0px',
            borderRadius=SpyderPalette.SIZE_BORDER_RADIUS
        )

        self.setStyleSheet(self.css.toString())

    def _start_spinner(self):
        """
        Start spinner when requested, in case the widget has one (it's stopped
        by default).
        """
        if self._spin is not None:
            self._spin.start()

    def _stop_spinner(self):
        """Stop spinner when requested, in case the widget has one."""
        if self._spin is not None:
            self._spin.stop()

    @qdebounced(timeout=30)
    def _on_resize_event(self):
        """Actions to take when widget is resized."""
        # Hide/show image label when necessary
        if self._image_label is not None:
            if (
                # This is necessary to prevent and error when the widget hasn't
                # been made visible yet.
                # Fixes spyder-ide/spyder#24280
                self._min_height is None
                # We need to do this validation because sometimes
                # minimumSizeHint doesn't give the right min_height in
                # showEvent (e.g. when adding an _ErroredMessageWidget to
                # plugins that are not visible).
                or self._min_height < self.minimumSizeHint().height()
            ):
                self._min_height = self.minimumSizeHint().height()

            if self.height() <= self._min_height:
                self._image_label.hide()
            else:
                if self._is_visible:
                    self._image_label.show()

        # Elide description when necessary
        if self._description is not None:
            metrics = QFontMetrics(self._description_label.font())

            # All margins are the same, so we only take the left one
            margin = self._description_label.contentsMargins().left()

            # Height of a single line of text
            text_line_height = metrics.height()

            # Allow a max of two lines of text in the description
            max_height = 2 * text_line_height + 2 * margin

            # Compute the width and height of the description text according to
            # max_height and the width of its label.
            # Solution taken from https://forum.qt.io/post/313343
            rect = metrics.boundingRect(
                # Rectangle in which the text needs to fit
                QRect(
                    0,
                    0,
                    self._description_label.width() - 2 * margin,
                    max_height,
                ),
                self._description_label.alignment() | Qt.TextWordWrap,
                self._description
            )

            # Elide text if it were to occupy more than two lines
            if rect.height() > 2 * text_line_height:
                # Replace description with elided text
                elided_text = metrics.elidedText(
                    self._description, Qt.ElideRight, rect.width()
                )
                self._description_label.setText(elided_text)

                # Show full description in tooltip
                self._description_label.setToolTip(
                    '\n'.join(textwrap.wrap(self._description, 50))
                )

                # This prevents flickering when the widget's width is
                # continuously reduced because Qt wraps the elided text
                self._description_label.setMaximumHeight(
                    text_line_height + 2 * margin
                )
            else:
                # Restore full description if there's enough space
                if "…" in self._description_label.text():
                    # Restore description
                    self._description_label.setText(self._description)

                    # No tooltip is necessary in this case
                    self._description_label.setToolTip("")

                    # Restore max height
                    self._description_label.setMaximumHeight(max_height)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Debugger Plugin."""

# Standard library imports
import os.path as osp
from typing import List

# Third-party imports
from qtpy.QtCore import Qt, Slot

# Local imports
from spyder.api.plugins import Plugins, SpyderDockablePlugin
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.shellconnect.mixins import ShellConnectPluginMixin
from spyder.api.translations import _
from spyder.plugins.debugger.confpage import DebuggerConfigPage
from spyder.plugins.debugger.utils.breakpointsmanager import (
    BreakpointsManager, clear_all_breakpoints, clear_breakpoint)
from spyder.plugins.debugger.widgets.main_widget import (
    DebuggerBreakpointActions, DebuggerWidget, DebuggerWidgetActions)
from spyder.plugins.editor.utils.editor import get_file_language
from spyder.plugins.editor.utils.languages import ALL_LANGUAGES
from spyder.plugins.ipythonconsole.api import IPythonConsolePyConfiguration
from spyder.plugins.mainmenu.api import ApplicationMenus, DebugMenuSections
from spyder.plugins.run.api import (
    RunConfiguration, ExtendedRunExecutionParameters, RunExecutor, run_execute,
    RunContext, RunResult)
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.plugins.ipythonconsole.widgets.run_conf import IPythonConfigOptions
from spyder.plugins.editor.api.run import CellRun, SelectionRun


class Debugger(SpyderDockablePlugin, ShellConnectPluginMixin, RunExecutor):
    """Debugger plugin."""

    NAME = 'debugger'
    REQUIRES = [Plugins.IPythonConsole, Plugins.Preferences, Plugins.Run]
    OPTIONAL = [Plugins.Editor, Plugins.MainMenu, Plugins.Toolbar]
    TABIFY = [Plugins.VariableExplorer, Plugins.Help]
    WIDGET_CLASS = DebuggerWidget
    CONF_SECTION = NAME
    CONF_FILE = False
    CONF_WIDGET_CLASS = DebuggerConfigPage
    DISABLE_ACTIONS_WHEN_HIDDEN = False

    # ---- SpyderDockablePlugin API
    # ------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _('Debugger')

    @staticmethod
    def get_description():
        return _('View, explore and navigate stack frames while debugging.')

    @classmethod
    def get_icon(cls):
        return cls.create_icon('debug')

    def on_initialize(self):
        widget = self.get_widget()
        widget.sig_pdb_state_changed.connect(
            self._update_current_codeeditor_pdb_state)
        widget.sig_toggle_breakpoints.connect(self._set_or_clear_breakpoint)
        widget.sig_toggle_conditional_breakpoints.connect(
            self._set_or_edit_conditional_breakpoint)
        widget.sig_clear_all_breakpoints.connect(self.clear_all_breakpoints)
        widget.sig_load_pdb_file.connect(self._load_pdb_file_in_editor)
        widget.sig_clear_breakpoint.connect(self.clear_breakpoint)
        widget.sig_switch_to_plugin_requested.connect(self.switch_to_plugin)

        self.python_editor_run_configuration = {
            'origin': self.NAME,
            'extension': 'py',
            'contexts': [
                {'name': 'File'},
                {'name': 'Cell'},
                {'name': 'Selection'},
            ]
        }

        self.ipython_editor_run_configuration = {
            'origin': self.NAME,
            'extension': 'ipy',
            'contexts': [
                {'name': 'File'},
                {'name': 'Cell'},
                {'name': 'Selection'},
            ]
        }

        self.pyw_editor_run_configuration = {
            'origin': self.NAME,
            'extension': 'pyw',
            'contexts': [
                {'name': 'File'},
                {'name': 'Cell'},
                {'name': 'Selection'},
            ]
        }

        self.executor_configuration = [
            {
                'input_extension': 'py',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': IPythonConfigOptions,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': 'ipy',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': IPythonConfigOptions,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': 'pyw',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': IPythonConfigOptions,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': 'py',
                'context': {'name': 'Cell'},
                'output_formats': [],
                'configuration_widget': None,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': 'ipy',
                'context': {'name': 'Cell'},
                'output_formats': [],
                'configuration_widget': None,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': 'pyw',
                'context': {'name': 'Cell'},
                'output_formats': [],
                'configuration_widget': None,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': 'py',
                'context': {'name': 'Selection'},
                'output_formats': [],
                'configuration_widget': None,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': 'ipy',
                'context': {'name': 'Selection'},
                'output_formats': [],
                'configuration_widget': None,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': 'pyw',
                'context': {'name': 'Selection'},
                'output_formats': [],
                'configuration_widget': None,
                'requires_cwd': True,
                'priority': 10
            },
        ]

    def on_mainwindow_visible(self):
        self.get_widget().update_splitter_widths(self.get_widget().width())

    @on_plugin_available(plugin=Plugins.Run)
    def on_run_available(self):
        run = self.get_plugin(Plugins.Run)
        run.register_executor_configuration(self, self.executor_configuration)

        run.create_run_in_executor_button(
            RunContext.File,
            self.NAME,
            text=_("&Debug file"),
            tip=_("Debug file"),
            icon=self.create_icon('debug'),
            shortcut_context="_",
            register_shortcut=True,
            add_to_menu={
                "menu": ApplicationMenus.Debug,
                "section": DebugMenuSections.StartDebug,
                "before_section": DebugMenuSections.ControlDebug
            },
            add_to_toolbar={
                "toolbar": ApplicationToolbars.Debug,
                "before": DebuggerWidgetActions.Next,
            },
            shortcut_widget_context=Qt.ApplicationShortcut,
        )

        run.create_run_in_executor_button(
            RunContext.Cell,
            self.NAME,
            text=_("Debug cell"),
            tip=_("Debug cell"),
            icon=self.create_icon('debug_cell'),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_menu={
                "menu": ApplicationMenus.Debug,
                "section": DebugMenuSections.StartDebug,
                "before_section": DebugMenuSections.ControlDebug
            },
            add_to_toolbar={
                "toolbar": ApplicationToolbars.Debug,
                "before": DebuggerWidgetActions.Next,
            },
        )

        run.create_run_in_executor_button(
            RunContext.Selection,
            self.NAME,
            text=_("Debug the current line or selection"),
            tip=_("Debug the current line or selection"),
            icon=self.create_icon('debug_selection'),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_menu={
                "menu": ApplicationMenus.Debug,
                "section": DebugMenuSections.StartDebug,
                "before_section": DebugMenuSections.ControlDebug
            },
            add_to_toolbar={
                "toolbar": ApplicationToolbars.Debug,
                "before": DebuggerWidgetActions.Next,
            },
        )

    @on_plugin_teardown(plugin=Plugins.Run)
    def on_run_teardown(self):
        run = self.get_plugin(Plugins.Run)
        run.deregister_executor_configuration(
            self, self.executor_configuration
        )
        run.destroy_run_in_executor_button(RunContext.File, self.NAME)
        run.destroy_run_in_executor_button(RunContext.Cell, self.NAME)
        run.destroy_run_in_executor_button(RunContext.Selection, self.NAME)

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @on_plugin_teardown(plugin=Plugins.Preferences)
    def on_preferences_teardown(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.deregister_plugin_preferences(self)

    @on_plugin_available(plugin=Plugins.Editor)
    def on_editor_available(self):
        editor = self.get_plugin(Plugins.Editor)
        widget = self.get_widget()

        for run_config in [
            self.python_editor_run_configuration,
            self.ipython_editor_run_configuration,
            self.pyw_editor_run_configuration
        ]:
            editor.add_supported_run_configuration(run_config)

        # The editor is available, connect signals.
        widget.sig_edit_goto.connect(editor.load)
        editor.sig_codeeditor_created.connect(self._add_codeeditor)
        editor.sig_codeeditor_changed.connect(self._update_codeeditor)
        editor.sig_codeeditor_deleted.connect(self._remove_codeeditor)

        # Apply shortcuts to editor and add actions to pythonfile list
        editor_shortcuts = [
            DebuggerBreakpointActions.ToggleBreakpoint,
            DebuggerBreakpointActions.ToggleConditionalBreakpoint,
            DebuggerBreakpointActions.ShowBreakpointsTable,
        ]
        for name in editor_shortcuts:
            action = self.get_action(name)
            # TODO: This should be handled differently?
            editor.get_widget().pythonfile_dependent_actions += [action]

    @on_plugin_teardown(plugin=Plugins.Editor)
    def on_editor_teardown(self):
        editor = self.get_plugin(Plugins.Editor)
        widget = self.get_widget()

        for run_config in [
            self.python_editor_run_configuration,
            self.ipython_editor_run_configuration,
            self.pyw_editor_run_configuration
        ]:
            editor.remove_supported_run_configuration(run_config)

        widget.sig_edit_goto.disconnect(editor.load)

        editor.sig_codeeditor_created.disconnect(self._add_codeeditor)
        editor.sig_codeeditor_changed.disconnect(self._update_codeeditor)
        editor.sig_codeeditor_deleted.disconnect(self._remove_codeeditor)

        # Remove editor actions
        editor_shortcuts = [
            DebuggerBreakpointActions.ToggleBreakpoint,
            DebuggerBreakpointActions.ToggleConditionalBreakpoint,
            DebuggerBreakpointActions.ShowBreakpointsTable,
        ]
        for name in editor_shortcuts:
            action = self.get_action(name)
            if action in editor.get_widget().pythonfile_dependent_actions:
                editor.get_widget().pythonfile_dependent_actions.remove(action)

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)

        # ControlDebug section
        for action in [DebuggerWidgetActions.Next,
                       DebuggerWidgetActions.Step,
                       DebuggerWidgetActions.Return,
                       DebuggerWidgetActions.Continue,
                       DebuggerWidgetActions.Stop]:
            mainmenu.add_item_to_application_menu(
                self.get_action(action),
                menu_id=ApplicationMenus.Debug,
                section=DebugMenuSections.ControlDebug,
                before_section=DebugMenuSections.EditBreakpoints)

        # Breakpoints section
        for action in [DebuggerBreakpointActions.ToggleBreakpoint,
                       DebuggerBreakpointActions.ToggleConditionalBreakpoint,
                       DebuggerBreakpointActions.ClearAllBreakpoints,
                       DebuggerBreakpointActions.ShowBreakpointsTable]:
            mainmenu.add_item_to_application_menu(
                self.get_action(action),
                menu_id=ApplicationMenus.Debug,
                section=DebugMenuSections.EditBreakpoints)

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        mainmenu = self.get_plugin(Plugins.MainMenu)

        names = [
            DebuggerWidgetActions.Next,
            DebuggerWidgetActions.Step,
            DebuggerWidgetActions.Return,
            DebuggerWidgetActions.Continue,
            DebuggerWidgetActions.Stop,
            DebuggerBreakpointActions.ToggleBreakpoint,
            DebuggerBreakpointActions.ToggleConditionalBreakpoint,
            DebuggerBreakpointActions.ClearAllBreakpoints,
            DebuggerBreakpointActions.ShowBreakpointsTable,
        ]
        for name in names:
            mainmenu.remove_item_from_application_menu(
                name,
                menu_id=ApplicationMenus.Debug
            )

    @on_plugin_available(plugin=Plugins.Toolbar)
    def on_toolbar_available(self):
        toolbar = self.get_plugin(Plugins.Toolbar)

        for action_id in [
            DebuggerWidgetActions.Next,
            DebuggerWidgetActions.Step,
            DebuggerWidgetActions.Return,
            DebuggerWidgetActions.Continue,
            DebuggerWidgetActions.Stop,
        ]:
            toolbar.add_item_to_application_toolbar(
                self.get_action(action_id),
                toolbar_id=ApplicationToolbars.Debug,
            )

        debug_toolbar = toolbar.get_application_toolbar(
            ApplicationToolbars.Debug
        )
        debug_toolbar.sig_is_rendered.connect(
            self.get_widget().on_debug_toolbar_rendered
        )

    @on_plugin_teardown(plugin=Plugins.Toolbar)
    def on_toolbar_teardown(self):
        toolbar = self.get_plugin(Plugins.Toolbar)

        for action_id in [
            DebuggerWidgetActions.Next,
            DebuggerWidgetActions.Step,
            DebuggerWidgetActions.Return,
            DebuggerWidgetActions.Continue,
            DebuggerWidgetActions.Stop,
        ]:
            toolbar.remove_item_from_application_toolbar(
                action_id,
                toolbar_id=ApplicationToolbars.Debug,
            )

    # ---- Private API
    # ------------------------------------------------------------------------
    def _load_pdb_file_in_editor(self, fname, lineno):
        """Load file using processevents."""
        editor = self.get_plugin(Plugins.Editor)
        if editor is None:
            return

        # Prevent keyboard input from accidentally entering the
        # editor during repeated, rapid entry of debugging commands.
        editor.load(fname, lineno, processevents=False)

    def _is_python_editor(self, codeeditor):
        """Check if the editor is a python editor."""
        if codeeditor.filename is None:
            return False
        txt = codeeditor.get_text_with_eol()
        language = get_file_language(codeeditor.filename, txt)
        return language.lower() in ALL_LANGUAGES["Python"]

    def _connect_codeeditor(self, codeeditor):
        """Connect a code editor."""
        codeeditor.breakpoints_manager = BreakpointsManager(codeeditor)
        codeeditor.breakpoints_manager.sig_breakpoints_saved.connect(
            self.get_widget().sig_breakpoints_saved)

    def _disconnect_codeeditor(self, codeeditor):
        """Connect a code editor."""
        codeeditor.breakpoints_manager.sig_breakpoints_saved.disconnect(
            self.get_widget().sig_breakpoints_saved)
        codeeditor.breakpoints_manager = None

    @Slot(str)
    def _filename_changed(self, filename):
        """Change filename."""
        codeeditor = self._get_editor_for_filename(filename)
        if codeeditor is None:
            return

        if codeeditor.breakpoints_manager is None:
            # Was not a python editor
            if self._is_python_editor(codeeditor):
                self._connect_codeeditor(codeeditor)
        else:
            # Was a python editor
            if self._is_python_editor(codeeditor):
                codeeditor.breakpoints_manager.set_filename(filename)
            else:
                self._disconnect_codeeditor(codeeditor)

    @Slot(object)
    def _add_codeeditor(self, codeeditor):
        """
        Add a new codeeditor.
        """
        codeeditor.sig_filename_changed.connect(self._filename_changed)
        codeeditor.breakpoints_manager = None
        if self._is_python_editor(codeeditor):
            self._connect_codeeditor(codeeditor)

    @Slot(object)
    def _remove_codeeditor(self, codeeditor):
        """
        Remove a codeeditor.
        """
        codeeditor.sig_filename_changed.disconnect(self._filename_changed)
        if codeeditor.breakpoints_manager is not None:
            self._disconnect_codeeditor(codeeditor)

    @Slot(object)
    def _update_codeeditor(self, codeeditor):
        """
        Focus codeeditor has changed.
        """
        if (
            codeeditor.filename is None or
            codeeditor.breakpoints_manager is None
        ):
            return

        # Update debugging state
        widget = self.get_widget()
        pdb_state = widget.get_pdb_state()
        filename, lineno = widget.get_pdb_last_step()
        codeeditor.breakpoints_manager.update_pdb_state(
            pdb_state, filename, lineno)

    @Slot(bool)
    def _update_current_codeeditor_pdb_state(self, pdb_state):
        """
        The pdb state has changed.
        """
        try:
            codeeditor = self._get_current_editor()
            if codeeditor is None or codeeditor.breakpoints_manager is None:
                return
            filename, line_number = self.get_widget().get_pdb_last_step()
            codeeditor.breakpoints_manager.update_pdb_state(
                pdb_state, filename, line_number)
        except RuntimeError:
            pass

    def _get_current_editor(self):
        """
        Get current codeeditor.
        """
        editor = self.get_plugin(Plugins.Editor)
        if editor is None:
            return None
        return editor.get_current_editor()

    def _get_editor_for_filename(self, filename):
        """Get editor for filename."""
        editor = self.get_plugin(Plugins.Editor)
        if editor is None:
            return None

        return editor.get_codeeditor_for_filename(filename)

    def _get_current_editorstack(self):
        """
        Get current editorstack.
        """
        editor = self.get_plugin(Plugins.Editor)
        if editor is None:
            return None
        return editor.get_current_editorstack()

    @Slot()
    def _set_or_clear_breakpoint(self):
        """Toggle breakpoint."""
        codeeditor = self._get_current_editor()
        if codeeditor is None or codeeditor.breakpoints_manager is None:
            return
        codeeditor.breakpoints_manager.toogle_breakpoint()

    @Slot()
    def _set_or_edit_conditional_breakpoint(self):
        """Set/edit conditional breakpoint."""
        codeeditor = self._get_current_editor()
        if codeeditor is None or codeeditor.breakpoints_manager is None:
            return
        codeeditor.breakpoints_manager.toogle_breakpoint(
            edit_condition=True)

    # ---- Public API
    # ------------------------------------------------------------------------

    @Slot()
    def clear_all_breakpoints(self):
        """Clear breakpoints in all files"""
        clear_all_breakpoints()
        self.get_widget().sig_breakpoints_saved.emit()

        editorstack = self._get_current_editorstack()
        if editorstack is not None:
            for data in editorstack.data:
                if data.editor.breakpoints_manager is not None:
                    data.editor.breakpoints_manager.clear_breakpoints()

    @Slot(str, int)
    def clear_breakpoint(self, filename, lineno):
        """Remove a single breakpoint"""
        clear_breakpoint(filename, lineno)
        self.get_widget().sig_breakpoints_saved.emit()

        codeeditor = self._get_editor_for_filename(filename)

        if codeeditor is None or codeeditor.breakpoints_manager is None:
            return None

        codeeditor.breakpoints_manager.toogle_breakpoint(lineno)

    def can_close_file(self, filename=None):
        """
        Check if a file can be closed taking into account debugging state.
        """
        if not self.get_conf('pdb_prevent_closing'):
            return True

        widget = self.get_widget()

        debugging = widget.get_pdb_state()
        if not debugging:
            return True

        pdb_fname, __ = widget.get_pdb_last_step()

        if pdb_fname and filename:
            if osp.normcase(pdb_fname) == osp.normcase(filename):
                widget.print_debug_file_msg()
                return False
            return True

        widget.print_debug_file_msg()
        return False

    # ---- For execution
    @run_execute(context=RunContext.File)
    def debug_file(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[RunResult]:

        console = self.get_plugin(Plugins.IPythonConsole)
        if console is None:
            return

        exec_params = conf['params']
        params: IPythonConsolePyConfiguration = exec_params['executor_params']
        params["run_method"] = "debugfile"

        console.exec_files(input, conf)

        self.get_widget().set_pdb_take_focus(False)

    @run_execute(context=RunContext.Cell)
    def debug_cell(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[RunResult]:

        console = self.get_plugin(Plugins.IPythonConsole)
        if console is None:
            return

        run_input: CellRun = input['run_input']
        if run_input['copy']:
            code = run_input['cell']
            if not code.strip():
                # Empty cell
                return
            console.run_selection("%%debug\n" + code)
            return

        exec_params = conf['params']
        params: IPythonConsolePyConfiguration = exec_params['executor_params']
        params["run_method"] = "debugcell"

        console.exec_cell(input, conf)

        self.get_widget().set_pdb_take_focus(False)


    @run_execute(context=RunContext.Selection)
    def debug_selection(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[RunResult]:

        console = self.get_plugin(Plugins.IPythonConsole)
        if console is None:
            return

        run_input: SelectionRun = input['run_input']
        code = run_input['selection']
        if not code.strip():
            # No selection
            return

        run_input['selection'] = "%%debug\n" + code

        console.exec_selection(input, conf)

        self.get_widget().set_pdb_take_focus(False)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
spyder.plugins.remoteclient.api.manager
=======================================

Remote Client Plugin API Manager.
"""

from __future__ import annotations

import asyncio
import logging
import socket
import typing
from abc import abstractmethod
from functools import partial

from spyder.api.asyncdispatcher import AsyncDispatcher
from spyder.api.translations import _
from spyder.api.utils import ABCMeta
from spyder.config.base import get_debug_level
from spyder.plugins.remoteclient.api.modules.base import (
    JupyterAPI,
)
from spyder.plugins.remoteclient.api.protocol import (
    ConnectionInfo,
    ConnectionStatus,
    RemoteClientLog,
)

if typing.TYPE_CHECKING:
    from spyder.plugins.remoteclient.api.modules.base import (
        SpyderBaseJupyterAPIType,
    )
    from spyder.plugins.remoteclient.plugin import RemoteClient


class SpyderRemoteAPILoggerHandler(logging.Handler):
    def __init__(self, client, *args, **kwargs):
        self._client = client
        super().__init__(*args, **kwargs)

        log_format = "%(message)s &#8212; %(asctime)s"
        formatter = logging.Formatter(log_format, datefmt="%H:%M:%S %d/%m/%Y")
        self.setFormatter(formatter)

    def emit(self, record):
        self._client._plugin.sig_client_message_logged.emit(
            RemoteClientLog(
                id=self._client.config_id,
                message=self.format(record),
                level=record.levelno,
                created=record.created,
            )
        )


class SpyderRemoteAPIManagerBase(metaclass=ABCMeta):
    """Class to manage a remote server and its APIs."""

    REGISTERED_MODULE_APIS: typing.ClassVar[
        dict[str, type[SpyderBaseJupyterAPIType]]
    ] = {
        JupyterAPI.__qualname__: JupyterAPI,
    }

    def __init__(
        self,
        conf_id: str,
        options: dict[str, typing.Any],
        _plugin: typing.Optional[RemoteClient] = None,
    ):
        self._config_id = conf_id
        self.options = options
        self._plugin = _plugin

        self.__create_events()

        self.__installing_server = False
        self.__starting_server = False
        self.__connection_task: asyncio.Task | None = None

        # For logging
        self.logger = logging.getLogger(
            f"{__name__}.{self.__class__.__name__}({self.config_id})"
        )

        if not get_debug_level():
            self.logger.setLevel(logging.DEBUG)

        if self._plugin is not None and not self.logger.hasHandlers():
            self.logger.addHandler(SpyderRemoteAPILoggerHandler(self))

    @AsyncDispatcher(loop="asyncssh", early_return=False)
    async def __create_events(self):
        self.__server_installed = asyncio.Event()
        self.__starting_event = asyncio.Event()
        self.__abort_requested = asyncio.Event()
        self.__lock = asyncio.Lock()

    def _emit_connection_status(self, status: str, message: str):
        if self._plugin is not None:
            self._plugin.sig_connection_status_changed.emit(
                ConnectionInfo(
                    id=self.config_id, status=status, message=message
                )
            )

    def _emit_version_mismatch(self, version: str):
        if self._plugin is not None:
            self._plugin.sig_version_mismatch.emit(self.config_id, version)

    @property
    def server_started(self):
        return self.__starting_event.is_set() and not self.__starting_server

    @property
    def connected(self) -> bool:
        """Check if a connection is currently established."""
        return (
            self.__connection_task is not None
            and self.__connection_task.done()
            and not self.__connection_task.cancelled()
            and not self.__connection_task.exception()
            and self.__connection_task.result() is True
        )

    @property
    def is_connecting(self) -> bool:
        """Check if a connection attempt is ongoing."""
        return (
            self.__connection_task is not None
            and not self.__connection_task.done()
        )

    @property
    def config_id(self):
        """Return the configuration ID"""
        return self._config_id

    @property
    def server_name(self):
        if self._plugin is None:
            return None
        return self._plugin.get_server_name(self.config_id)

    @property
    @abstractmethod
    def server_url(self):
        """Return Jypyter server URL."""
        msg = "This method should be implemented in the derived class"
        raise NotImplementedError(msg)

    @property
    @abstractmethod
    def api_token(self):
        """Return Jupyter server API token."""
        msg = "This method should be implemented in the derived class"
        raise NotImplementedError(msg)

    async def close(self):
        """Closes the remote server and the SSH connection."""
        self._emit_connection_status(
            ConnectionStatus.Stopping,
            _("We're closing the connection. Please be patient"),
        )

        await self.stop_remote_server()
        await self.close_connection()

    def _handle_connection_lost(self, exc: Exception | None = None):
        self._reset_connection_established()
        self.__starting_event.clear()
        self.logger.error(
            "Connection to %s was lost",
            self.server_name,
            exc_info=exc,
        )
        self._emit_connection_status(
            status=ConnectionStatus.Error,
            message=_("The connection was lost"),
        )
        if self._plugin:
            self._plugin.sig_connection_lost.emit(self.config_id)

    # ---- Connection and server management
    async def connect_and_install_remote_server(self) -> bool:
        """Connect to the remote server and install the server."""
        if await self.create_new_connection():
            return await self.install_remote_server()

        return False

    async def connect_and_start_server(self) -> bool:
        """Connect to the remote server and start the server."""
        if await self.create_new_connection():
            return await self.start_remote_server()

        return False

    async def ensure_connection_and_server(self) -> bool:
        """Ensure connected and the remote server is running."""
        if self.connected and not self.server_started:
            return await self.ensure_server()

        if not self.connected:
            return await self.connect_and_ensure_server()

        return True

    async def connect_and_ensure_server(self) -> bool:
        """Connect and ensure the remote server is running."""
        if await self.create_new_connection() and not self.server_started:
            return await self.ensure_server()

        return bool(self.server_started)

    async def ensure_connection(self) -> bool:
        """Ensure the SSH connection is open."""
        if self.connected:
            return True

        return await self.create_new_connection()

    async def ensure_server(self, *, check_installed: bool = True) -> bool:
        """Ensure remote server is installed and running."""
        if self.server_started:
            return True

        if check_installed and not await self.ensure_server_installed():
            return False

        return await self.start_remote_server()

    async def start_remote_server(self):
        """Start remote server."""
        if self.__starting_server:
            await self.__starting_event.wait()
            return self.server_started

        self.__starting_server = True
        self._emit_connection_status(
            ConnectionStatus.Starting, _("Starting Spyder remote services")
        )
        try:
            if await self._start_remote_server():
                self.__starting_event.set()
                # emit signal that connection and server are established
                if self._plugin:
                    self._plugin.sig_connection_established.emit(
                        self.config_id
                    )
                return True
        finally:
            self.__starting_server = False

        self.__starting_event.clear()
        return False

    @abstractmethod
    async def _start_remote_server(self):
        """Start remote server."""
        msg = "This method should be implemented in the derived class"
        raise NotImplementedError(msg)

    @abstractmethod
    async def ensure_server_installed(self) -> bool:
        """Check remote server version."""
        msg = "This method should be implemented in the derived class"
        raise NotImplementedError(msg)

    async def install_remote_server(self) -> bool:
        """Install remote server."""
        if self.__installing_server:
            await self.__server_installed.wait()
            return True

        self.__installing_server = True
        try:
            if await self._install_remote_server():
                self.__server_installed.set()
                return True
        finally:
            self.__installing_server = False

        self.__server_installed.clear()
        return False

    @abstractmethod
    async def _install_remote_server(self):
        """Install remote server."""
        msg = "This method should be implemented in the derived class"
        raise NotImplementedError(msg)

    async def create_new_connection(self) -> bool:
        """Create a new connection, supporting user abort."""
        if self.connected:
            self.logger.debug(
                "Atempting to create a new connection with an existing for %s",
                self.server_name,
            )
            await self.close_connection()

        async with self.__lock:
            if self.__connection_task is not None:
                return self.connected

            self._emit_connection_status(
                ConnectionStatus.Connecting,
                _("We're establishing the connection. Please be patient"),
            )

            self.__abort_requested.clear()
            self.__connection_task = asyncio.create_task(
                self._create_new_connection()
            )
            abort_task = asyncio.create_task(self.__abort_requested.wait())

            await asyncio.wait(
                [self.__connection_task, abort_task],
                return_when=asyncio.FIRST_COMPLETED,
            )

            # User aborted
            if self.__abort_requested.is_set():
                self.__connection_task.cancel()
                abort_task.cancel()
                self.__connection_task = None

                self.logger.debug("Connection attempt aborted by user")
                self._emit_connection_status(
                    ConnectionStatus.Inactive,
                    _("The connection attempt was cancelled"),
                )

                return False

            try:
                # Connection completed
                if await self.__connection_task:
                    self._emit_connection_status(
                        ConnectionStatus.Connected,
                        _("The connection was successfully established"),
                    )
                    return True
            except BaseException as error:
                # Log any error
                self.logger.error(
                    "Error creating a new connection for {}. The error was:"
                    "<br>{}".format(self.server_name, str(error))
                )

            # Cancel and reset connection tasks to allow retries after errors
            self.__connection_task.cancel()
            self.__connection_task = None
            abort_task.cancel()

            # Report that the connection failed
            self._emit_connection_status(
                ConnectionStatus.Error,
                _("There was an error establishing the connection"),
            )

            return False

    async def abort_connection(self):
        """Abort an ongoing connection attempt."""
        if self.is_connecting:
            self.__abort_requested.set()

    @abstractmethod
    async def _create_new_connection(self) -> bool:
        """Create a new connection to the remote server machine."""
        msg = "This method should be implemented in the derived class"
        raise NotImplementedError(msg)

    async def stop_remote_server(self):
        """Close remote server."""
        if not self.server_started:
            self.logger.warning(
                f"Remote server is not running for {self.server_name}"
            )
            return False

        if not self.connected:
            self.logger.error("SSH connection is not open")
            self._emit_connection_status(
                ConnectionStatus.Error,
                _("The SSH connection is not open"),
            )
            return False

        # bug in jupyterhub, need to send SIGINT twice
        self.logger.debug(f"Stopping remote server for {self.server_name}")
        try:
            async with self.get_jupyter_api() as jupyter:
                await jupyter.shutdown_server()
        except Exception as err:
            self.logger.error(
                "Error stopping remote server. The error was:<br>{}".format(
                    str(err)
                )
            )

        self.__starting_event.clear()
        self.logger.info(
            f"Remote server process closed for {self.server_name}"
        )
        return True

    @abstractmethod
    async def close_connection(self):
        """Close SSH connection."""
        msg = "This method should be implemented in the derived class"
        raise NotImplementedError(msg)

    def _reset_connection_established(self):
        """Reset the connection status."""
        self.__connection_task = None

    @staticmethod
    def get_free_port():
        """Request a free port from the OS."""
        with socket.socket() as s:
            s.bind(("", 0))
            return s.getsockname()[1]

    # ---- API Management
    @staticmethod
    def register_api(
        kclass: type[SpyderBaseJupyterAPIType]
    ) -> type[SpyderBaseJupyterAPIType]:
        """Register a REST API class."""
        SpyderRemoteAPIManagerBase.REGISTERED_MODULE_APIS[
            kclass.__qualname__
        ] = kclass
        return kclass

    def get_api(
        self,
        api: str | type[SpyderBaseJupyterAPIType]
    ) -> typing.Callable[..., SpyderBaseJupyterAPIType]:
        """Get a registered REST API class."""
        if isinstance(api, type):
            api = api.__qualname__

        api_class = SpyderRemoteAPIManagerBase.REGISTERED_MODULE_APIS.get(api)
        if api_class is None:
            raise ValueError(f"API {api} is not registered")

        return partial(api_class, manager=self)

    def get_jupyter_api(self) -> JupyterAPI:
        return JupyterAPI(manager=self)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Kernel handler."""

# Standard library imports
import json
import os
import os.path as osp
from subprocess import PIPE
from threading import Lock
import uuid

# Third-party imports
from jupyter_core.paths import jupyter_runtime_dir
from qtpy.QtCore import QObject, QThread, Signal, Slot

# Local imports
from spyder.api.translations import _
from spyder.config.base import running_under_pytest
from spyder.plugins.ipythonconsole import (
    SPYDER_KERNELS_MIN_VERSION,
    SPYDER_KERNELS_MAX_VERSION,
    SPYDER_KERNELS_VERSION,
    SPYDER_KERNELS_CONDA,
    SPYDER_KERNELS_PIP,
    SpyderKernelError,
)
from spyder.plugins.ipythonconsole.comms.kernelcomm import KernelComm
from spyder.plugins.ipythonconsole.utils.manager import SpyderKernelManager
from spyder.plugins.ipythonconsole.utils.client import SpyderKernelClient
from spyder.plugins.ipythonconsole.utils.websocket_client import (
    SpyderWSKernelClient,
)
from spyder.utils.programs import check_version_range


PERMISSION_ERROR_MSG = _(
    "The directory <tt>{}</tt> is not writable and it is required to create "
    "IPython consoles. Please make it writable."
)

ERROR_SPYDER_KERNEL_VERSION = _(
    "The Python environment or installation whose interpreter is located at"
    "<pre>"
    "    <tt>{0}</tt>"
    "</pre>"
    "doesn't have the right version of <tt>spyder-kernels</tt> installed ({1} "
    "instead of >= {2} and < {3}). Without this module is not possible for "
    "Spyder to create a console for you.<br><br>"
    "You can install it by activating your environment (if necessary) and "
    "then running in a system terminal:"
    "<pre>"
    "    <tt>{4}</tt>"
    "</pre>"
    "or"
    "<pre>"
    "    <tt>{5}</tt>"
    "</pre>"
)

# For Spyder-kernels version < 3.0, where the version and executable cannot be queried
ERROR_SPYDER_KERNEL_VERSION_OLD = _(
    "This Python environment doesn't have the right version of "
    "<tt>spyder-kernels</tt> installed (>= {0} and < {1}). Without this "
    "module is not possible for Spyder to create a console for you.<br><br>"
    "You can install it by activating your environment (if necessary) and "
    "then running in a system terminal:"
    "<pre>"
    "    <tt>{2}</tt>"
    "</pre>"
    "or"
    "<pre>"
    "    <tt>{3}</tt>"
    "</pre>"
)


class KernelConnectionState:
    SpyderKernelWaitComm = "spyder_kernel_wait_comm"
    SpyderKernelReady = "spyder_kernel_ready"
    IpykernelReady = "ipykernel_ready"
    Connecting = "connecting"
    Error = "error"
    Closed = "closed"
    Crashed = "crashed"


class StdThread(QThread):
    """Poll for changes in std buffers."""

    sig_out = Signal(str)

    def __init__(self, parent, std_buffer):
        super().__init__(parent)
        self._std_buffer = std_buffer
        self._closing = False

    def run(self):
        txt = True
        while txt:
            try:
                txt = self._std_buffer.read1()
            except ValueError:  # I/O operation on closed file
                break

            if txt:
                try:
                    txt = txt.decode()
                except UnicodeDecodeError:
                    txt = str(txt)
                self.sig_out.emit(txt)
            else:
                break  # EOF


class KernelHandler(QObject):
    """
    A class to handle the kernel in several ways and store kernel connection
    information.
    """

    sig_stdout = Signal(str)
    """
    A stdout message was received on the process stdout.
    """

    sig_stderr = Signal(str)
    """
    A stderr message was received on the process stderr.
    """

    sig_fault = Signal(str)
    """
    A fault message was received.
    """

    sig_kernel_is_ready = Signal()
    """
    The kernel is ready.
    """

    sig_kernel_connection_error = Signal()
    """
    The kernel raised an error while connecting.
    """

    _shutdown_thread_list = []
    """List of running shutdown threads"""

    _shutdown_thread_list_lock = Lock()
    """
    Lock to add threads to _shutdown_thread_list or clear that list.
    """

    def __init__(
        self,
        kernel_client,
        connection_file=None,
        kernel_manager=None,
        known_spyder_kernel=False,
        hostname=None,
        sshkey=None,
        password=None,
        ssh_connection=None,
        websocket_url=None,
        token=None,
        aiohttp_session=None,
    ):
        super().__init__()
        # Connection Informations
        self.connection_file = connection_file
        self.kernel_manager = kernel_manager
        self.kernel_client = kernel_client
        self.known_spyder_kernel = known_spyder_kernel

        self.hostname = hostname
        self.sshkey = sshkey
        self.password = password

        self.ssh_connection = ssh_connection

        self.websocket_url = websocket_url
        self.token = token

        self.aiohttp_session = None

        self.kernel_error_message = None
        self.connection_state = KernelConnectionState.Connecting

        # Comm
        self.kernel_comm = KernelComm()
        self.kernel_comm.sig_comm_ready.connect(self.handle_comm_ready)

        # Internal
        self._shutdown_lock = Lock()
        self._stdout_thread = None
        self._stderr_thread = None
        self._fault_args = None
        self._init_stderr = ""
        self._init_stdout = ""
        self._shellwidget_connected = False
        self._comm_ready_received = False

        # Start kernel
        self.kernel_client.sig_spyder_kernel_info.connect(
            self.check_spyder_kernel_info
        )
        self.connect_std_pipes()
        self.kernel_client.start_channels()

        # Open comm and wait for comm ready reply.
        # It only works for spyder-kernels, but this is the majority of cases.
        # For ipykernels, this does nothing.
        self.kernel_comm.open_comm(self.kernel_client)

    @property
    def is_websocket_client(self):
        """Return the websocket client."""
        return isinstance(self.kernel_client, SpyderWSKernelClient)

    def connect_(self):
        """Connect to shellwidget."""
        self._shellwidget_connected = True
        # Emit signal in case the connection is already made
        if self.connection_state in [
            KernelConnectionState.IpykernelReady,
            KernelConnectionState.SpyderKernelReady
        ]:
            # This is necessary for systems in which the kernel takes too much
            # time to start because in that case its heartbeat is not detected
            # as beating at this point.
            # Fixes spyder-ide/spyder#22179
            self.kernel_client.hb_channel._beating = True
            self.sig_kernel_is_ready.emit()
        elif self.connection_state == KernelConnectionState.Error:
            self.sig_kernel_connection_error.emit()

        # Show initial io
        if self._init_stderr:
            self.sig_stderr.emit(self._init_stderr)
        self._init_stderr = None
        if self._init_stdout:
            self.sig_stdout.emit(self._init_stdout)
        self._init_stdout = None

    def check_spyder_kernel_info(self, spyder_kernel_info):
        """
        Check if the Spyder-kernels version is the right one after receiving it
        from the kernel.

        If the kernel is non-locally managed, check if it is a spyder-kernel.
        """

        if not spyder_kernel_info:
            if self.known_spyder_kernel:
                # spyder-kernels version < 3.0
                self.kernel_error_message = (
                    ERROR_SPYDER_KERNEL_VERSION_OLD.format(
                        SPYDER_KERNELS_MIN_VERSION,
                        SPYDER_KERNELS_MAX_VERSION,
                        SPYDER_KERNELS_CONDA,
                        SPYDER_KERNELS_PIP,
                    )
                )
                self.connection_state = KernelConnectionState.Error
                self.known_spyder_kernel = False
                self.sig_kernel_connection_error.emit()
                return

            self.connection_state = KernelConnectionState.IpykernelReady
            self.sig_kernel_is_ready.emit()
            return

        version, pyexec = spyder_kernel_info
        if not check_version_range(version, SPYDER_KERNELS_VERSION):
            # Development versions are acceptable
            if "dev0" not in version:
                self.kernel_error_message = ERROR_SPYDER_KERNEL_VERSION.format(
                    pyexec,
                    version,
                    SPYDER_KERNELS_MIN_VERSION,
                    SPYDER_KERNELS_MAX_VERSION,
                    SPYDER_KERNELS_CONDA,
                    SPYDER_KERNELS_PIP,
                )
                self.known_spyder_kernel = False
                self.connection_state = KernelConnectionState.Error
                self.sig_kernel_connection_error.emit()
                return

        self.known_spyder_kernel = True
        self.connection_state = KernelConnectionState.SpyderKernelWaitComm
        if self._comm_ready_received:
            self.handle_comm_ready()

    def handle_comm_ready(self):
        """The kernel comm is ready"""
        self._comm_ready_received = True
        if self.connection_state in [
            KernelConnectionState.SpyderKernelWaitComm,
            KernelConnectionState.Crashed,
        ]:
            # This is necessary for systems in which the kernel takes too much
            # time to start because in that case its heartbeat is not detected
            # as beating at this point.
            # Fixes spyder-ide/spyder#22179
            self.kernel_client.hb_channel._beating = True
            self.connection_state = KernelConnectionState.SpyderKernelReady
            self.sig_kernel_is_ready.emit()

    def connect_std_pipes(self):
        """Connect to std pipes."""
        self.close_std_threads()

        # Connect new threads
        if self.kernel_manager is None:
            return

        stdout = self.kernel_manager.provisioner.process.stdout
        stderr = self.kernel_manager.provisioner.process.stderr

        if stdout:
            self._stdout_thread = StdThread(self, stdout)
            self._stdout_thread.sig_out.connect(self.handle_stdout)
            self._stdout_thread.start()
        if stderr:
            self._stderr_thread = StdThread(self, stderr)
            self._stderr_thread.sig_out.connect(self.handle_stderr)
            self._stderr_thread.start()

    def disconnect_std_pipes(self):
        """Disconnect old std pipes."""
        if self._stdout_thread and not self._stdout_thread._closing:
            self._stdout_thread.sig_out.disconnect(self.handle_stdout)
            self._stdout_thread._closing = True
        if self._stderr_thread and not self._stderr_thread._closing:
            self._stderr_thread.sig_out.disconnect(self.handle_stderr)
            self._stderr_thread._closing = True

    def close_std_threads(self):
        """Close std threads."""
        if self._stdout_thread is not None:
            self._stdout_thread.wait()
            self._stdout_thread = None
        if self._stderr_thread is not None:
            self._stderr_thread.wait()
            self._stderr_thread = None

    @Slot(str)
    def handle_stderr(self, err):
        """Handle stderr"""
        if self._shellwidget_connected:
            self.sig_stderr.emit(err)
        else:
            self._init_stderr += err

    @Slot(str)
    def handle_stdout(self, out):
        """Handle stdout"""
        if self._shellwidget_connected:
            self.sig_stdout.emit(out)
        else:
            self._init_stdout += out

    @staticmethod
    def new_connection_file():
        """
        Generate a new connection file

        Taken from jupyter_client/console_app.py
        Licensed under the BSD license
        """
        # Check if jupyter_runtime_dir exists (Spyder addition)
        if not osp.isdir(jupyter_runtime_dir()):
            try:
                os.makedirs(jupyter_runtime_dir())
            except (IOError, OSError):
                return None
        cf = ""
        while not cf:
            ident = str(uuid.uuid4()).split("-")[-1]
            cf = os.path.join(jupyter_runtime_dir(), "kernel-%s.json" % ident)
            cf = cf if not os.path.exists(cf) else ""
        return cf

    @classmethod
    def new_from_spec(cls, kernel_spec):
        """
        Create a new kernel.

        Might raise all kinds of exceptions
        """
        connection_file = cls.new_connection_file()
        if connection_file is None:
            raise SpyderKernelError(
                PERMISSION_ERROR_MSG.format(jupyter_runtime_dir())
            )

        # Kernel manager
        kernel_manager = SpyderKernelManager(
            connection_file=connection_file,
            config=None,
            autorestart=True,
        )

        kernel_manager._kernel_spec = kernel_spec

        try:
            kernel_manager.start_kernel(
                stderr=PIPE,
                stdout=PIPE,
                env=kernel_spec.env,
            )
        except PermissionError:
            # Show a nice error message when jupyter_runtime_dir is not
            # writable.
            # Fixes spyder-ide/spyder#23124
            raise SpyderKernelError(
                PERMISSION_ERROR_MSG.format(jupyter_runtime_dir())
            )

        # Kernel client
        kernel_client = kernel_manager.client()

        # Increase time (in seconds) to detect if a kernel is alive.
        # See spyder-ide/spyder#3444.
        kernel_client.hb_channel.time_to_dead = 25.0

        return cls(
            connection_file=connection_file,
            kernel_manager=kernel_manager,
            kernel_client=kernel_client,
            known_spyder_kernel=True,
        )

    @classmethod
    def from_connection_info(
        cls,
        connection_info,
        hostname=None,
        sshkey=None,
        password=None,
        ssh_connection=None,
    ):
        """Create kernel for given connection info."""
        new_connection_file = cls.new_connection_file()
        with open(new_connection_file, "w") as f:
            json.dump(connection_info, f)

        return cls(
            connection_file=new_connection_file,
            kernel_client=cls.init_kernel_client(
                new_connection_file,
                hostname,
                sshkey,
                password,
                ssh_connection,
            ),
        )

    @classmethod
    def from_connection_file(
        cls,
        connection_file,
        hostname=None,
        sshkey=None,
        password=None,
        ssh_connection=None,
    ):
        """Create kernel for given connection file."""
        return cls(
            connection_file=connection_file,
            hostname=hostname,
            sshkey=sshkey,
            password=password,
            kernel_client=cls.init_kernel_client(
                connection_file,
                hostname,
                sshkey,
                password,
                ssh_connection,
            ),
        )

    @classmethod
    def from_websocket(
        cls,
        websocket_url,
        token=None,
        aiohttp_session=None,
    ):
        return cls(
            websocket_url=websocket_url,
            token=token,
            aiohttp_session=aiohttp_session,
            kernel_client=cls.init_ws_kernel_client(
                websocket_url,
                token=token,
                aiohttp_session=aiohttp_session,
            ),
        )

    @staticmethod
    def init_kernel_client(
        connection_file,
        hostname,
        sshkey,
        password,
        ssh_connection,
    ):
        """Create kernel client."""
        kernel_client = SpyderKernelClient(connection_file=connection_file)

        # This is needed for issue spyder-ide/spyder#9304.
        try:
            kernel_client.load_connection_file()
        except Exception as e:
            raise SpyderKernelError(
                _(
                    "An error occurred while trying to load "
                    "the kernel connection file. The error "
                    "was:\n\n"
                )
                + f"<tt>{str(e)}</tt>"
            )

        if hostname is not None or ssh_connection is not None:
            kernel_client.tunnel_to_kernel(
                hostname=hostname,
                sshkey=sshkey,
                password=password,
                ssh_connection=ssh_connection,
            )

        return kernel_client

    @staticmethod
    def init_ws_kernel_client(
        websocket_url,
        token=None,
        username=None,
        aiohttp_session=None,
    ):
        """Create kernel client."""
        return SpyderWSKernelClient(
            endpoint=websocket_url,
            token=token,
            username=username,
            aiohttp_session=aiohttp_session,
        )

    def close(self, shutdown_kernel=True, now=False):
        """Close kernel"""
        self.close_comm()
        if shutdown_kernel and self.kernel_manager is not None:
            km = self.kernel_manager
            km.stop_restarter()
            self.disconnect_std_pipes()

            # This is probably necessary due to a weird interaction between
            # `conda run --no-capture-output` and pytest capturing output
            # facilities.
            if now or running_under_pytest():
                km.shutdown_kernel(now=True)
                self.after_shutdown()
            else:
                shutdown_thread = QThread(None)
                shutdown_thread.run = self._thread_shutdown_kernel
                shutdown_thread.finished.connect(self.after_shutdown)
                shutdown_thread.start()
                with self._shutdown_thread_list_lock:
                    self._shutdown_thread_list.append(shutdown_thread)

        if (
            self.kernel_client is not None
            and self.kernel_client.channels_running
        ):
            self.kernel_client.stop_channels()

    def after_shutdown(self):
        """Cleanup after shutdown"""
        self.close_std_threads()
        self.kernel_comm.remove(only_closing=True)

    def _thread_shutdown_kernel(self):
        """Shutdown kernel."""
        with self._shutdown_lock:
            # Avoid calling shutdown_kernel on the same manager twice
            # from different threads to avoid crash.
            if self.kernel_manager.shutting_down:
                return
            self.kernel_manager.shutting_down = True
        try:
            self.kernel_manager.shutdown_kernel()
        except Exception:
            # kernel was externally killed
            pass

    @classmethod
    def wait_all_shutdown_threads(cls):
        """Wait shutdown thread."""
        with cls._shutdown_thread_list_lock:
            for thread in cls._shutdown_thread_list:
                if thread.isRunning():
                    try:
                        thread.kernel_manager._kill_kernel()
                    except Exception:
                        pass
                    thread.quit()
                    thread.wait()
            cls._shutdown_thread_list = []

    def copy(self):
        """Copy kernel."""
        # Copy kernel infos

        # Get new kernel_client
        if self.is_websocket_client:
            kernel_client = self.init_ws_kernel_client(
                self.websocket_url,
                token=self.token,
                aiohttp_session=self.aiohttp_session,
            )
        else:
            kernel_client = self.init_kernel_client(
                self.connection_file,
                self.hostname,
                self.sshkey,
                self.password,
                self.ssh_connection,
            )

        return self.__class__(
            connection_file=self.connection_file,
            kernel_manager=self.kernel_manager,
            known_spyder_kernel=self.known_spyder_kernel,
            hostname=self.hostname,
            sshkey=self.sshkey,
            password=self.password,
            ssh_connection=self.ssh_connection,
            websocket_url=self.websocket_url,
            token=self.token,
            aiohttp_session=self.aiohttp_session,
            kernel_client=kernel_client,
        )

    def faulthandler_setup(self, args):
        """Setup faulthandler"""
        self._fault_args = args

    def poll_fault_text(self):
        """Get a fault from a previous session."""
        if self._fault_args is None:
            return
        self.kernel_comm.remote_call(
            callback=self.emit_fault_text
        ).get_fault_text(*self._fault_args)
        self._fault_args = None

    def emit_fault_text(self, fault):
        """Emit fault text"""
        if not fault:
            return
        self.sig_fault.emit(fault)

    def fault_filename(self):
        """Get fault filename."""
        if not self._fault_args:
            return
        return self._fault_args[0]

    def close_comm(self):
        """Close comm"""
        self.connection_state = KernelConnectionState.Closed
        self.kernel_comm.close()

    def reopen_comm(self):
        """Reopen comm (following a crash)"""
        self.kernel_comm.remove()
        self.connection_state = KernelConnectionState.Crashed
        self.kernel_comm.open_comm(self.kernel_client)

    def reconnect_kernel(self):
        """Kernel restarted successfully, so reconnect to it."""
        # Unecessary closing and opening comms as it's already restarted
        # in `ShellWidget._handle_kernel_restarted` with
        # `self.kernel_handler.reopen_comm`. But this is required for the
        # `sig_kernel_is_ready` to be emitted when running our tests.
        # TODO: Check this logic again
        self.kernel_comm.close()
        self.connection_state = KernelConnectionState.Crashed
        self.kernel_comm.open_comm(self.kernel_client)

    def set_time_to_dead(self, time_to_dead):
        """Set time to detect if the kernel is dead in seconds."""
        if self.kernel_client is not None:
            self.kernel_client.hb_channel.time_to_dead = time_to_dead
